#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import time
import uuid
import os

# Sleep constants
from parlai.mturk.core import data_model as data_model

THREAD_SHORT_SLEEP = 0.1
THREAD_MEDIUM_SLEEP = 0.3
# ThrottlingException might happen if we poll too frequently
THREAD_MTURK_POLLING_SLEEP = 10

logger = None
logging_enabled = True
debug = False
log_level = logging.ERROR

if logging_enabled:
    logging.basicConfig(
        filename=str(time.time()) + '.log',
        filemode='w',
        format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
        datefmt='%H:%M:%S',
        level=logging.DEBUG,
    )
    logger = logging.getLogger('mturk')


def set_log_level(new_level):
    global log_level
    log_level = new_level


def set_is_debug(is_debug):
    global debug
    debug = is_debug


def disable_logging():
    global logging_enabled
    logging_enabled = False


def print_and_log(level, message, should_print=False):
    if (logging_enabled and level >= log_level) or debug:
        logger.log(level, message)
    if should_print or debug:  # Always print message in debug mode
        print(message)


def generate_event_id(worker_id):
    """
    Return a unique id to use for identifying a packet for a worker.
    """
    return '{}_{}'.format(worker_id, uuid.uuid4())


def get_mturk_dir():
    import parlai.mturk

    return os.path.dirname(os.path.abspath(parlai.mturk.__file__))


def get_core_dir():
    import parlai.mturk.core

    return os.path.dirname(os.path.abspath(parlai.mturk.core.__file__))


def get_tmp_dir():
    """
    Return the location of the temporary directory in which we store things related to a
    run but that can be safely deleted.
    """
    tmp_dir = os.path.join(get_mturk_dir(), 'tmp')
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
    return tmp_dir


class AssignState:
    """
    Class for holding state information about an assignment currently claimed by an
    agent.
    """

    # Possible Assignment Status Values
    STATUS_NONE = 'none'
    STATUS_ONBOARDING = 'onboarding'
    STATUS_WAITING = 'waiting'
    STATUS_IN_TASK = 'in task'
    STATUS_DONE = 'done'
    STATUS_DISCONNECT = 'disconnect'
    STATUS_PARTNER_DISCONNECT = 'partner disconnect'
    STATUS_PARTNER_DISCONNECT_EARLY = 'partner disconnect early'
    STATUS_EXPIRED = 'expired'
    STATUS_RETURNED = 'returned'

    def __init__(self, status=None):
        """
        Create an AssignState to track the state of an agent's assignment.
        """
        if status is None:
            status = self.STATUS_NONE
        self.status = status
        self.messages = []
        self.last_command = None
        self.message_ids = []

    def clear_messages(self):
        self.messages = []
        self.message_ids = []
        self.last_command = None

    def append_message(self, message):
        """
        Appends a message to the list of messages, ensures that it is not a duplicate
        message.
        """
        if message['message_id'] in self.message_ids:
            return
        self.message_ids.append(message['message_id'])
        self.messages.append(message)

    def set_last_command(self, command):
        self.last_command = command

    def get_last_command(self):
        return self.last_command

    def get_messages(self):
        return self.messages

    def set_status(self, status):
        """
        Set the status of this agent on the task.
        """
        # TODO log to db
        self.status = status

    def get_status(self):
        """
        Get the status of this agent on its task.
        """
        # TODO retrieve from db if not set
        return self.status

    def is_final(self):
        """
        Return True if the assignment is in a final status that can no longer be acted
        on.
        """
        return (
            self.status == self.STATUS_DISCONNECT
            or self.status == self.STATUS_DONE
            or self.status == self.STATUS_PARTNER_DISCONNECT
            or self.status == self.STATUS_PARTNER_DISCONNECT_EARLY
            or self.status == self.STATUS_RETURNED
            or self.status == self.STATUS_EXPIRED
        )

    def get_inactive_command_text(self):
        """
        Get appropriate inactive command and text to respond to a reconnect given the
        current assignment state.

        returns text, command
        """
        command = data_model.COMMAND_INACTIVE_HIT
        text = None
        if self.status == self.STATUS_DISCONNECT:
            text = (
                'You disconnected in the middle of this HIT and were '
                'marked as inactive. As these HITs often require real-'
                'time interaction, it is no longer available for '
                'completion. Please return this HIT and accept a new one '
                'if you would like to try again.'
            )
        elif self.status == self.STATUS_DONE:
            command = data_model.COMMAND_INACTIVE_DONE
            text = (
                'You disconnected after completing this HIT without '
                'marking it as completed. Please press the done button '
                'below to finish the HIT.'
            )
        elif self.status == self.STATUS_EXPIRED:
            text = (
                'You disconnected in the middle of this HIT and the '
                'HIT expired before you reconnected. It is no longer '
                'available for completion. Please return this HIT and '
                'accept a new one if you would like to try again.'
            )
        elif self.status == self.STATUS_PARTNER_DISCONNECT:
            command = data_model.COMMAND_INACTIVE_DONE
            text = (
                'One of your partners disconnected in the middle of the '
                'HIT. We won\'t penalize you for their disconnect, so '
                'please use the button below to mark the HIT as complete.'
            )
        elif self.status == self.STATUS_PARTNER_DISCONNECT_EARLY:
            command = data_model.COMMAND_INACTIVE_HIT
            text = (
                'One of your partners disconnected in the middle of the '
                'HIT. We won\'t penalize you for their disconnect, but you'
                ' did not complete enough of the task to submit the HIT. '
                'Please return this HIT and accept a new one if you would '
                'like to try again.'
            )
        elif self.status == self.STATUS_RETURNED:
            text = (
                'You disconnected from this HIT and then returned '
                'it. As we have marked the HIT as returned, it is no '
                'longer available for completion. Please accept a new '
                'HIT if you would like to try again'
            )
        else:
            # We shouldn't be getting an inactive command for the other
            # states so consider this a server error
            text = (
                'Our server was unable to handle your reconnect properly '
                'and thus this HIT no longer seems available for '
                'completion. Please try to connect again or return this '
                'HIT and accept a new one.'
            )

        return text, command



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BART Module.
"""
import torch
import torch.nn.functional as F
from typing import Any, Dict, Union, List, Optional
from parlai.agents.transformer.modules import TransformerGeneratorModel


class BartModel(TransformerGeneratorModel):
    """
    BART Model.
    """

    def output(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Compute output logits.

        Override standard TGM output to _not_ prevent generation of BOS.
        """
        # project back to vocabulary
        output = F.linear(tensor, self.embeddings.weight)
        return output

    def _get_initial_forced_decoder_input(self, bsz: int, inputs: torch.LongTensor):
        """
        Return initial input to the decoder.

        Override TGA._get_initial_forced_decoder_input to seed EOS BOS.

        :param bsz:
            batchsize
        :param inputs:
            inputs to decode

        :return initial_input:
            initial input for the decoder.
        """
        tens = (
            torch.LongTensor([self.END_IDX, self.START_IDX])
            .to(inputs)
            .detach()
            .expand(bsz, 2)
        )
        return torch.cat([tens, inputs], 1)

    def reorder_decoder_incremental_state(
        self,
        incremental_state: Dict[str, Any],
        inds: Union[List[int], torch.LongTensor],
    ) -> Optional[Dict[str, Any]]:
        """
        Incremental state is weird to handle when we seed decoder with two inputs
        initially.

        We leave as a future exercise.
        """
        return None



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import time
from queue import Queue
import uuid

from parlai.core.agents import Agent
import parlai.mturk.core.data_model as data_model
import parlai.mturk.core.shared_utils as shared_utils

# Special act messages for failure states
MTURK_DISCONNECT_MESSAGE = '[DISCONNECT]'  # Turker disconnected from conv
TIMEOUT_MESSAGE = '[TIMEOUT]'  # the Turker did not respond but didn't return
RETURN_MESSAGE = '[RETURNED]'  # the Turker returned the HIT

# TODO move time management into another class, this way we can handle it
# relative to heartbeats. This will come with more thorough testing.


class AssignState:
    """
    Class for holding state information about an assignment currently claimed by an
    agent.
    """

    # Possible Assignment Status Values
    STATUS_NONE = 'none'
    STATUS_ONBOARDING = 'onboarding'
    STATUS_WAITING = 'waiting'
    STATUS_IN_TASK = 'in task'
    STATUS_DONE = 'done'
    STATUS_DISCONNECT = 'disconnect'
    STATUS_PARTNER_DISCONNECT = 'partner disconnect'
    STATUS_PARTNER_DISCONNECT_EARLY = 'partner disconnect early'
    STATUS_EXPIRED = 'expired'
    STATUS_RETURNED = 'returned'

    def __init__(self, status=None):
        """
        Create an AssignState to track the state of an agent's assignment.
        """
        if status is None:
            status = self.STATUS_NONE
        self.status = status
        self.messages = []
        self.last_command = None
        self.message_ids = []

    def clear_messages(self):
        self.messages = []
        self.message_ids = []
        self.last_command = None

    def append_message(self, message):
        """
        Appends a message to the list of messages, ensures that it is not a duplicate
        message.
        """
        if message['message_id'] in self.message_ids:
            return
        self.message_ids.append(message['message_id'])
        self.messages.append(message)

    def set_last_command(self, command):
        self.last_command = command

    def get_last_command(self):
        return self.last_command

    def get_messages(self):
        return self.messages

    def set_status(self, status):
        """
        Set the status of this agent on the task.
        """
        # TODO log to db
        self.status = status

    def get_status(self):
        """
        Get the status of this agent on its task.
        """
        # TODO retrieve from db if not set
        return self.status

    def is_final(self):
        """
        Return True if the assignment is in a final status that can no longer be acted
        on.
        """
        return (
            self.status == self.STATUS_DISCONNECT
            or self.status == self.STATUS_DONE
            or self.status == self.STATUS_PARTNER_DISCONNECT
            or self.status == self.STATUS_PARTNER_DISCONNECT_EARLY
            or self.status == self.STATUS_RETURNED
            or self.status == self.STATUS_EXPIRED
        )

    def get_inactive_command_text(self):
        """
        Get appropriate inactive command and text to respond to a reconnect given the
        current assignment state.

        returns text, command
        """
        command = data_model.COMMAND_INACTIVE_HIT
        text = None
        if self.status == self.STATUS_DISCONNECT:
            text = (
                'You disconnected in the middle of this HIT and were '
                'marked as inactive. As these HITs often require real-'
                'time interaction, it is no longer available for '
                'completion. Please return this HIT and accept a new one '
                'if you would like to try again.'
            )
        elif self.status == self.STATUS_DONE:
            command = data_model.COMMAND_INACTIVE_DONE
            text = (
                'You disconnected after completing this HIT without '
                'marking it as completed. Please press the done button '
                'below to finish the HIT.'
            )
        elif self.status == self.STATUS_EXPIRED:
            text = (
                'You disconnected in the middle of this HIT and the '
                'HIT expired before you reconnected. It is no longer '
                'available for completion. Please return this HIT and '
                'accept a new one if you would like to try again.'
            )
        elif self.status == self.STATUS_PARTNER_DISCONNECT:
            command = data_model.COMMAND_INACTIVE_DONE
            text = (
                'One of your partners disconnected in the middle of the '
                'HIT. We won\'t penalize you for their disconnect, so '
                'please use the button below to mark the HIT as complete.'
            )
        elif self.status == self.STATUS_PARTNER_DISCONNECT_EARLY:
            command = data_model.COMMAND_INACTIVE_HIT
            text = (
                'One of your partners disconnected in the middle of the '
                'HIT. We won\'t penalize you for their disconnect, but you'
                ' did not complete enough of the task to submit the HIT. '
                'Please return this HIT and accept a new one if you would '
                'like to try again.'
            )
        elif self.status == self.STATUS_RETURNED:
            text = (
                'You disconnected from this HIT and then returned '
                'it. As we have marked the HIT as returned, it is no '
                'longer available for completion. Please accept a new '
                'HIT if you would like to try again'
            )
        else:
            # We shouldn't be getting an inactive command for the other
            # states so consider this a server error
            text = (
                'Our server was unable to handle your reconnect properly '
                'and thus this HIT no longer seems available for '
                'completion. Please try to connect again or return this '
                'HIT and accept a new one.'
            )

        return text, command


class MTurkAgent(Agent):
    """
    Base class for an MTurkAgent that can act in a ParlAI world.
    """

    # MTurkAgent Possible Statuses
    ASSIGNMENT_NOT_DONE = 'NotDone'
    ASSIGNMENT_DONE = 'Submitted'
    ASSIGNMENT_APPROVED = 'Approved'
    ASSIGNMENT_REJECTED = 'Rejected'

    MTURK_DISCONNECT_MESSAGE = MTURK_DISCONNECT_MESSAGE
    TIMEOUT_MESSAGE = TIMEOUT_MESSAGE
    RETURN_MESSAGE = RETURN_MESSAGE

    def __init__(self, opt, mturk_manager, hit_id, assignment_id, worker_id):
        super().__init__(opt)

        self.conversation_id = None
        self.mturk_manager = mturk_manager
        self.db_logger = mturk_manager.db_logger
        self.id = None
        self.state = AssignState()
        self.assignment_id = assignment_id
        self.hit_id = hit_id
        self.worker_id = worker_id
        self.some_agent_disconnected = False
        self.hit_is_expired = False
        self.hit_is_abandoned = False  # state from Amazon MTurk system
        self.hit_is_returned = False  # state from Amazon MTurk system
        self.hit_is_complete = False  # state from Amazon MTurk system
        self.disconnected = False
        self.task_group_id = mturk_manager.task_group_id
        self.message_request_time = None
        self.recieved_packets = {}
        self.creation_time = time.time()
        self.alived = True  # Used for restoring state after refresh
        self.feedback = None

        self.msg_queue = Queue()

    def _get_episode_done_msg(self, text):
        return {'id': self.id, 'text': text, 'episode_done': True}

    def set_status(self, status):
        """
        Set the status of this agent on the task, update db.
        """
        self.state.set_status(status)
        if self.db_logger is not None:
            if status == AssignState.STATUS_ONBOARDING:
                self.db_logger.log_start_onboard(
                    self.worker_id, self.assignment_id, self.conversation_id
                )
            elif status == AssignState.STATUS_WAITING:
                self.db_logger.log_finish_onboard(self.worker_id, self.assignment_id)
            elif status == AssignState.STATUS_IN_TASK:
                self.db_logger.log_start_task(
                    self.worker_id, self.assignment_id, self.conversation_id
                )
            elif status == AssignState.STATUS_DONE:
                self.db_logger.log_complete_assignment(
                    self.worker_id,
                    self.assignment_id,
                    time.time() + self.mturk_manager.auto_approve_delay,
                    status,
                )
            elif status == AssignState.STATUS_PARTNER_DISCONNECT:
                self.db_logger.log_complete_assignment(
                    self.worker_id,
                    self.assignment_id,
                    time.time() + self.mturk_manager.auto_approve_delay,
                    status,
                )
            elif status == AssignState.STATUS_PARTNER_DISCONNECT_EARLY:
                self.db_logger.log_complete_assignment(
                    self.worker_id,
                    self.assignment_id,
                    time.time() + self.mturk_manager.auto_approve_delay,
                    status,
                )
            elif status == AssignState.STATUS_DISCONNECT:
                self.db_logger.log_disconnect_assignment(
                    self.worker_id,
                    self.assignment_id,
                    time.time() + self.mturk_manager.auto_approve_delay,
                    status,
                )
            elif status == AssignState.STATUS_EXPIRED:
                self.db_logger.log_complete_assignment(
                    self.worker_id,
                    self.assignment_id,
                    time.time() + self.mturk_manager.auto_approve_delay,
                    status,
                )
            elif status == AssignState.STATUS_RETURNED:
                self.db_logger.log_abandon_assignment(
                    self.worker_id, self.assignment_id
                )

    def get_status(self):
        """
        Get the status of this agent on its task.
        """
        return self.state.get_status()

    def submitted_hit(self):
        return self.get_status() in [
            AssignState.STATUS_DONE,
            AssignState.STATUS_PARTNER_DISCONNECT,
        ]

    def is_final(self):
        """
        Determine if this agent is in a final state.
        """
        return self.state.is_final()

    def append_message(self, message):
        """
        Add a received message to the state.
        """
        self.state.append_message(message)

    def set_last_command(self, command):
        """
        Changes the last command recorded as sent to the agent.
        """
        self.state.set_last_command(command)

    def get_last_command(self):
        """
        Returns the last command to be sent to this agent.
        """
        return self.state.get_last_command()

    def clear_messages(self):
        """
        Clears the message history for this agent.
        """
        self.state.clear_messages()

    def get_messages(self):
        """
        Returns all the messages stored in the state.
        """
        return self.state.get_messages()

    def get_connection_id(self):
        """
        Returns an appropriate connection_id for this agent.
        """
        return "{}_{}".format(self.worker_id, self.assignment_id)

    def log_reconnect(self):
        """
        Log a reconnect of this agent.
        """
        shared_utils.print_and_log(
            logging.DEBUG,
            'Agent ({})_({}) reconnected to {} with status {}'.format(
                self.worker_id,
                self.assignment_id,
                self.conversation_id,
                self.get_status(),
            ),
        )

    def get_inactive_command_data(self):
        """
        Get appropriate inactive command data to respond to a reconnect.
        """
        text, command = self.state.get_inactive_command_text()
        return {
            'text': command,
            'inactive_text': text,
            'conversation_id': self.conversation_id,
            'agent_id': self.worker_id,
        }

    def wait_for_status(self, desired_status):
        """
        Suspend a thread until a particular assignment state changes to the desired
        state.
        """
        while True:
            if self.get_status() == desired_status:
                return True
            if self.is_final():
                return False
            time.sleep(shared_utils.THREAD_SHORT_SLEEP)

    def is_in_task(self):
        """
        Use conversation_id to determine if an agent is in a task.
        """
        if self.conversation_id:
            return 't_' in self.conversation_id
        return False

    def observe(self, msg):
        """
        Send an agent a message through the mturk_manager.
        """
        self.mturk_manager.send_message(self.worker_id, self.assignment_id, msg)

    def put_data(self, id, data):
        """
        Put data into the message queue if it hasn't already been seen.
        """
        if id not in self.recieved_packets:
            self.recieved_packets[id] = True
            self.msg_queue.put(data)

    def flush_msg_queue(self):
        """
        Clear all messages in the message queue.

        Return flushed messages
        """
        messages = []
        if self.msg_queue is None:
            return []
        while not self.msg_queue.empty():
            messages.append(self.msg_queue.get())
        return messages

    def reduce_state(self):
        """
        Cleans up resources related to maintaining complete state.
        """
        self.flush_msg_queue()
        self.msg_queue = None
        self.recieved_packets = None

    def get_new_act_message(self):
        """
        Get a new act message if one exists, return None otherwise.
        """
        # See if any agent has disconnected
        if self.disconnected or self.some_agent_disconnected:
            return self._get_episode_done_msg(MTURK_DISCONNECT_MESSAGE)

        # Check if the current turker already returned the HIT
        if self.hit_is_returned:
            return self._get_episode_done_msg(RETURN_MESSAGE)

        if self.msg_queue is not None:
            # Check if Turker sends a message
            while not self.msg_queue.empty():
                msg = self.msg_queue.get()
                if msg['id'] == self.id:
                    return msg

        # There are no messages to be sent
        return None

    def prepare_timeout(self):
        """
        Log a timeout event, tell mturk manager it occurred, return message to return
        for the act call.
        """
        shared_utils.print_and_log(
            logging.INFO, '{} timed out before sending.'.format(self.id)
        )
        self.mturk_manager.handle_turker_timeout(self.worker_id, self.assignment_id)
        return self._get_episode_done_msg(TIMEOUT_MESSAGE)

    def request_message(self):
        if not (
            self.disconnected or self.some_agent_disconnected or self.hit_is_expired
        ):
            self.mturk_manager.send_command(
                self.worker_id,
                self.assignment_id,
                {'text': data_model.COMMAND_SEND_MESSAGE},
            )

    def act(self, timeout=None, blocking=True):
        """
        Sends a message to other agents in the world.

        If blocking, this will wait for the message to come in so it can be sent.
        Otherwise it will return None.
        """
        if not blocking:
            # if this is the first act since last sent message start timing
            if self.message_request_time is None:
                self.request_message()
                self.message_request_time = time.time()

            # If checking timeouts
            if timeout:
                # If time is exceeded, timeout
                if time.time() - self.message_request_time > timeout:
                    return self.prepare_timeout()

            # Get a new message, if it's not None reset the timeout
            msg = self.get_new_act_message()
            if msg is not None and self.message_request_time is not None:
                self.message_request_time = None
            return msg
        else:
            self.request_message()
            self.message_request_time = time.time()

            # Timeout in seconds, after which the HIT is expired automatically
            if timeout:
                start_time = time.time()

            # Wait for agent's new message
            while True:
                msg = self.get_new_act_message()
                self.message_request_time = None
                if msg is not None:
                    return msg

                # Check if the Turker waited too long to respond
                if timeout:
                    current_time = time.time()
                    if (current_time - start_time) > timeout:
                        self.message_request_time = None
                        return self.prepare_timeout()
                time.sleep(shared_utils.THREAD_SHORT_SLEEP)

    def episode_done(self):
        """
        Return whether or not this agent believes the conversation to be done.
        """
        if (
            self.mturk_manager.get_agent_work_status(self.assignment_id)
            == self.ASSIGNMENT_NOT_DONE
        ):
            return False
        else:
            return True

    def _print_not_available_for(self, item):
        shared_utils.print_and_log(
            logging.WARN,
            'Conversation ID: {}, Agent ID: {} - HIT '
            'is abandoned and thus not available for '
            '{}.'.format(self.conversation_id, self.id, item),
            should_print=True,
        )

    def approve_work(self):
        """
        Approving work after it has been submitted.
        """
        if self.hit_is_abandoned:
            self._print_not_available_for('review')
        else:
            if (
                self.mturk_manager.get_agent_work_status(self.assignment_id)
                == self.ASSIGNMENT_DONE
            ):
                self.mturk_manager.approve_work(assignment_id=self.assignment_id)
                shared_utils.print_and_log(
                    logging.INFO,
                    'Conversation ID: {}, Agent ID: {} - HIT is '
                    'approved.'.format(self.conversation_id, self.id),
                )
            else:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Cannot approve HIT. Turker hasn\'t completed the HIT yet.',
                )

    def reject_work(self, reason='unspecified'):
        """
        Reject work after it has been submitted.
        """
        if self.hit_is_abandoned:
            self._print_not_available_for('review')
        else:
            if (
                self.mturk_manager.get_agent_work_status(self.assignment_id)
                == self.ASSIGNMENT_DONE
            ):
                self.mturk_manager.reject_work(self.assignment_id, reason)
                shared_utils.print_and_log(
                    logging.INFO,
                    'Conversation ID: {}, Agent ID: {} - HIT is '
                    'rejected.'.format(self.conversation_id, self.id),
                )
            else:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Cannot reject HIT. Turker hasn\'t completed the HIT yet.',
                )

    def block_worker(self, reason='unspecified'):
        """
        Block a worker from our tasks.
        """
        self.mturk_manager.block_worker(worker_id=self.worker_id, reason=reason)
        shared_utils.print_and_log(
            logging.WARN,
            'Blocked worker ID: {}. Reason: {}'.format(self.worker_id, reason),
            should_print=True,
        )

    def pay_bonus(self, bonus_amount, reason='unspecified'):
        """
        Pays the given agent the given bonus.
        """
        if self.hit_is_abandoned:
            self._print_not_available_for('bonus')
        else:
            if self.mturk_manager.get_agent_work_status(self.assignment_id) in (
                self.ASSIGNMENT_DONE,
                self.ASSIGNMENT_APPROVED,
            ):
                unique_request_token = str(uuid.uuid4())
                self.mturk_manager.pay_bonus(
                    worker_id=self.worker_id,
                    bonus_amount=bonus_amount,
                    assignment_id=self.assignment_id,
                    reason=reason,
                    unique_request_token=unique_request_token,
                )
            else:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Cannot pay bonus for HIT. Reason: Turker '
                    'hasn\'t completed the HIT yet.',
                )

    def email_worker(self, subject, message_text):
        """
        Sends an email to a worker, returns true on a successful send.
        """
        response = self.mturk_manager.email_worker(
            worker_id=self.worker_id, subject=subject, message_text=message_text
        )
        if 'success' in response:
            shared_utils.print_and_log(
                logging.INFO,
                'Email sent to worker ID: {}: Subject: {}: Text: {}'.format(
                    self.worker_id, subject, message_text
                ),
            )
            return True
        elif 'failure' in response:
            shared_utils.print_and_log(
                logging.WARN,
                "Unable to send email to worker ID: {}. Error: {}".format(
                    self.worker_id, response['failure']
                ),
            )
            return False

    def set_hit_is_abandoned(self):
        """
        Update local state to abandoned and mark the HIT as expired.
        """
        if not self.hit_is_abandoned:
            self.hit_is_abandoned = True
            self.mturk_manager.force_expire_hit(self.worker_id, self.assignment_id)

    def wait_completion_timeout(self, iterations):
        """
        Suspends the thread waiting for hit completion for some number of iterations on
        the THREAD_MTURK_POLLING_SLEEP time.
        """

        # Determine number of sleep iterations for the amount of time
        # we want to wait before syncing with MTurk. Start with 10 seconds
        # of waiting
        iters = (
            shared_utils.THREAD_MTURK_POLLING_SLEEP / shared_utils.THREAD_MEDIUM_SLEEP
        )
        i = 0
        # Wait for the desired number of MTURK_POLLING_SLEEP iterations
        while not self.hit_is_complete and i < iters * iterations:
            time.sleep(shared_utils.THREAD_SHORT_SLEEP)
            i += 1
        return

    def wait_for_hit_completion(self, timeout=None):
        """
        Waits for a hit to be marked as complete.
        """
        # Timeout in seconds, after which the HIT will be expired automatically
        if timeout:
            if timeout < 0:
                # Negative timeout is for testing, wait for packet to send
                time.sleep(1)
                self.mturk_manager.free_workers([self])
                return True
            start_time = time.time()
        wait_periods = 1
        self.wait_completion_timeout(wait_periods)
        sync_attempts = 0
        while (
            not self.hit_is_complete
            and self.mturk_manager.get_agent_work_status(self.assignment_id)
            != self.ASSIGNMENT_DONE
        ):
            if sync_attempts < 8:
                # Scaling on how frequently to poll, doubles time waited on
                # every failure
                wait_periods *= 2
                sync_attempts += 1
            else:
                # Okay we've waited for 45 mins and the HIT still isn't up
                self.disconnected = True
            # Check if the Turker already returned/disconnected
            if self.hit_is_returned or self.disconnected:
                self.mturk_manager.free_workers([self])
                return False
            if timeout:
                current_time = time.time()
                if (current_time - start_time) > timeout:
                    shared_utils.print_and_log(
                        logging.INFO,
                        "Timeout waiting for ({})_({}) to complete {}.".format(
                            self.worker_id, self.assignment_id, self.conversation_id
                        ),
                    )
                    self.set_hit_is_abandoned()
                    self.mturk_manager.free_workers([self])
                    return False
            shared_utils.print_and_log(
                logging.DEBUG,
                'Waiting for ({})_({}) to complete {}...'.format(
                    self.worker_id, self.assignment_id, self.conversation_id
                ),
            )
            self.wait_completion_timeout(wait_periods)

        shared_utils.print_and_log(
            logging.INFO,
            'Conversation ID: {}, Agent ID: {} - HIT is done.'.format(
                self.conversation_id, self.id
            ),
        )
        self.mturk_manager.free_workers([self])
        return True

    def shutdown(self, timeout=None, direct_submit=False):
        """
        Shuts down a hit when it is completed.
        """
        # Timeout in seconds, after which the HIT will be expired automatically
        command_to_send = data_model.COMMAND_SHOW_DONE_BUTTON
        if direct_submit:
            command_to_send = data_model.COMMAND_SUBMIT_HIT
        if not (
            self.hit_is_abandoned
            or self.hit_is_returned
            or self.disconnected
            or self.hit_is_expired
        ):
            self.mturk_manager.mark_workers_done([self])
            self.mturk_manager.send_command(
                self.worker_id, self.assignment_id, {'text': command_to_send}
            )
            did_complete = self.wait_for_hit_completion(timeout=timeout)
            if did_complete and self.db_logger is not None:
                self.db_logger.log_submit_assignment(self.worker_id, self.assignment_id)
            # Grab feedback message if it happens to exist
            messages = self.flush_msg_queue()
            for m in messages:
                if m['text'] == '[PEER_REVIEW]':
                    self.feedback = m['task_data']
            return did_complete

    def update_agent_id(self, agent_id):
        """
        Workaround used to force an update to an agent_id on the front-end to render the
        correct react components for onboarding and waiting worlds.

        Only really used in special circumstances where different agents need different
        onboarding worlds.
        """
        self.mturk_manager.worker_manager.change_agent_conversation(
            agent=self, conversation_id=self.conversation_id, new_agent_id=agent_id
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
MockTorchAgent.

Mean for unit testing purposes only, and should not be invoked otherwise.
"""

from parlai.core.torch_agent import TorchAgent, Output
import torch
from parlai.core.agents import Agent


class MockDict(Agent):
    """
    Mock Dictionary Agent which just implements indexing and txt2vec.
    """

    null_token = '__null__'
    NULL_IDX = 0
    start_token = '__start__'
    BEG_IDX = 1001
    end_token = '__end__'
    END_IDX = 1002
    p1_token = '__p1__'
    P1_IDX = 2001
    p2_token = '__p2__'
    P2_IDX = 2002

    def __init__(self, opt, shared=None):
        """
        Initialize idx for incremental indexing.
        """
        self.idx = 0

    def __getitem__(self, key):
        """
        Return index of special token or return the token.
        """
        if key == self.null_token:
            return self.NULL_IDX
        elif key == self.start_token:
            return self.BEG_IDX
        elif key == self.end_token:
            return self.END_IDX
        elif key == self.p1_token:
            return self.P1_IDX
        elif key == self.p2_token:
            return self.P2_IDX
        else:
            self.idx += 1
            return self.idx

    def __setitem__(self, key, value):
        pass

    def __len__(self):
        return 0

    def add_cmdline_args(self, *args, **kwargs):
        """
        Add CLI args.
        """
        pass

    def txt2vec(self, txt):
        """
        Return index of special tokens or range from 1 for each token.
        """
        self.idx = 0
        return [self[tok] for tok in txt.split()]

    def save(self, path, sort=False):
        """
        Override to do nothing.
        """
        pass


class MockTorchAgent(TorchAgent):
    """
    Use MockDict instead of regular DictionaryAgent.
    """

    @staticmethod
    def dictionary_class():
        """
        Replace normal dictionary class with mock one.
        """
        return MockDict

    def __init__(self, opt, shared=None):
        self.model = self.build_model()
        self.criterion = self.build_criterion()
        super().__init__(opt, shared)

    def build_model(self):
        return torch.nn.Module()

    def build_criterion(self):
        return torch.nn.NLLLoss()

    def train_step(self, batch):
        """
        Return confirmation of training.
        """
        return Output(['Training {}!'.format(i) for i in range(len(batch.text_vec))])

    def eval_step(self, batch):
        """
        Return confirmation of evaluation.
        """
        return Output(
            [
                'Evaluating {} (responding to {})!'.format(
                    i, batch.observations[i]['text']
                )
                for i in range(len(batch.text_vec))
            ]
        )


class SilentTorchAgent(TorchAgent):
    """
    Use MockDict instead of regular DictionaryAgent.
    """

    @staticmethod
    def dictionary_class():
        """
        Replace normal dictionary class with mock one.
        """
        return MockDict

    def __init__(self, opt, shared=None):
        self.model = self.build_model()
        self.criterion = self.build_criterion()
        super().__init__(opt, shared)

    def build_model(self):
        return torch.nn.Module()

    def build_criterion(self):
        return torch.nn.NLLLoss()

    def train_step(self, batch):
        """
        Null output.
        """
        return Output()

    def eval_step(self, batch):
        """
        Null output.
        """
        return Output()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Dict, List, Set, Any
import json
import os
import queue
import random
import time

from parlai.core.params import ParlaiParser
from parlai.mturk.core.mturk_manager import StaticMTurkManager
from parlai.mturk.core.worlds import StaticMTurkTaskWorld
from parlai.utils.misc import warn_once


DEFAULT_TASK_CONFIG = {
    'hit_title': 'Which Conversational Partner is Better?',
    'hit_description': 'Evaluate quality of conversations through comparison.',
    'hit_keywords': 'chat,evaluation,comparison,conversation',
}

AGENT_DISPLAY_NAME = 'RatingWorker'


def add_args(from_argv=False):
    """
    Add arguments to parser and either parse from commandline or initialize to defaults
    (for overriding in scripts)
    """
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    argparser.add_argument(
        '--annotations-per-pair',
        type=int,
        default=1,
        help='Number of annotations per conversation comparison pair',
    )
    argparser.add_argument(
        '--pairings-filepath',
        type=str,
        default=None,
        help='path to the file containing the task dictionaries',
    )
    argparser.add_argument(
        '--task-config',
        type=dict,
        default=DEFAULT_TASK_CONFIG,
        help='dict with keys "hit_title", "hit_description", "hit_keywords", '
        'determining how task is displayed on MTurk site',
    )
    argparser.add_argument(
        '--s1-choice',
        type=str,
        default='I would prefer to talk to <Speaker 1>',
        help='text next to speaker 1 radio button',
    )
    argparser.add_argument(
        '--s2-choice',
        type=str,
        default='I would prefer to talk to <Speaker 2>',
        help='text next to speaker 2 radio button',
    )
    argparser.add_argument(
        '--question',
        type=str,
        default='Who would you prefer to talk to for a long conversation?',
        help='question to present to turker for comparison (e.g. "Which speaker is better?")',
    )
    argparser.add_argument(
        '--block-on-onboarding-fail',
        type=bool,
        default=True,
        help='whether to block on onboarding failure',
    )
    argparser.add_argument(
        '--subtasks-per-hit',
        type=int,
        default=5,
        help='number of subtasks/comparisons to do per hit',
    )
    argparser.add_argument(
        '--onboarding-threshold',
        type=float,
        default=0.75,
        help='minimum accuracy on onboarding tasks, as a float 0-1.0',
    )
    argparser.add_argument('--seed', type=int, default=42, help='seed for random')
    argparser.add_argument(
        '--softblock-list-path',
        type=str,
        default=None,
        help='Path to list of workers to softblock, separated by line breaks',
    )
    argparser.set_defaults(allowed_conversation=1)
    if from_argv:
        return argparser.parse_args()
    else:
        return argparser.parse_args(args=[])


class AcuteEvaluator(object):
    """
    Run ACUTE Eval.

    Relevant args are parsed in the `setup_args` function above.
    """

    def __init__(self, opt):
        """
        Initialize the AcuteEvaluator.

        The following object attributes are used in running ACUTE Eval:

        ``onboarding_tasks``: A list of ALL available _onboarding_ comparison tasks

        ``desired_tasks``: A list of ALL available comparison tasks

        ``task_queue``: A queue of REMAINING tasks, from which HITs are constructed.

        ``worker_data``: A mapping from worker ID to data about the worker, including
        their tasks completed, conversations seen, and onboarding todo

        ``failed_onboard``:   The set of workers who have failed onboarding
        """
        random.seed(opt['seed'])
        self.opt = opt

        # add additional opt args
        self._supplement_opt()

        # class attributes
        self.onboarding_tasks: List[Dict] = []
        self.desired_tasks: List[Dict] = []
        self.task_queue: queue.Queue = queue.Queue()
        self.worker_data: Dict[str, Dict[str, List]] = {}
        self.failed_onboard: Set = set()

        # read in conversations data
        self._load_conversation_data()

        # setup the task queue
        self._setup_task_queue()

        # instantiate Manager
        self.manager = StaticMTurkManager(opt=self.opt)

    def _get_worker_data(self, worker_id: str) -> Dict[str, List]:
        """
        Return worker data if present, else a default dict.
        """
        onboarding_todo = list(range(len(self.onboarding_tasks)))
        random.shuffle(onboarding_todo)
        self.worker_data[worker_id] = self.worker_data.get(
            worker_id,
            {
                'tasks_completed': [],
                'conversations_seen': [],
                'onboarding_todo': onboarding_todo,
            },
        )
        return self.worker_data[worker_id]

    def _supplement_opt(self):
        """
        Add additional args to opt.

        Useful to add relevant options after args are parsed.
        """
        self.opt.update(
            {
                'task': os.path.basename(os.path.dirname(os.path.abspath(__file__))),
                'task_description': {
                    'num_subtasks': self.opt['subtasks_per_hit'],
                    'question': self.opt['question'],
                },
                'frontend_version': 1,
            }
        )
        self.opt.update(self.opt['task_config'])

    def set_block_qual(self, task_group_id: str):
        """
        Set block qualification if necessary.

        :param task_group_id:
            task id used to set block qualification, if necessary.
        """
        if (
            self.opt['block_on_onboarding_fail']
            and self.opt['block_qualification'] is None
        ):
            self.opt['block_qualification'] = task_group_id
            warn_once(
                "No block_qualification set in opt, automatically creating "
                "new qualification {}".format(task_group_id)
            )

    def _load_conversation_data(self):
        """
        Load conversation data.

        Loads in the data from the pairs filepath.
        """
        pairs_path = self.opt.get('pairings_filepath')
        if not os.path.exists(pairs_path):
            raise RuntimeError('You MUST specify a valid pairings filepath')

        with open(pairs_path) as pf:
            for i, l in enumerate(pf.readlines()):
                convo_pair = json.loads(l.strip())
                eval_speakers = [
                    s
                    for d in convo_pair['dialogue_dicts']
                    for s in d['speakers']
                    if s in convo_pair['speakers_to_eval']
                ]
                # make sure order is preserved
                assert eval_speakers == convo_pair['speakers_to_eval']
                model_left_idx = random.choice([0, 1])
                task = {
                    'task_specs': {
                        's1_choice': self.opt['s1_choice'],
                        's2_choice': self.opt['s2_choice'],
                        'question': self.opt['question'],
                        'is_onboarding': convo_pair['is_onboarding'],
                        'model_left': {
                            'name': eval_speakers[model_left_idx],
                            'dialogue': convo_pair['dialogue_dicts'][model_left_idx][
                                'dialogue'
                            ],
                        },
                        'model_right': {
                            'name': eval_speakers[1 - model_left_idx],
                            'dialogue': convo_pair['dialogue_dicts'][
                                1 - model_left_idx
                            ]['dialogue'],
                        },
                    },
                    'pairing_dict': convo_pair,
                    'pair_id': i,
                }
                if convo_pair.get('is_onboarding'):
                    self.onboarding_tasks.append(task)
                else:
                    self.desired_tasks.append(task)

    def _setup_task_queue(self):
        """
        Fill task queue with conversation pairs.
        """
        for _i in range(self.opt['annotations_per_pair']):
            all_task_keys = list(range(len(self.desired_tasks)))
            random.shuffle(all_task_keys)
            for p_id in all_task_keys:
                self.task_queue.put(self.desired_tasks[p_id])

    def _get_dialogue_ids(self, task: Dict[str, Any]) -> List[int]:
        """
        Return the ids for the dialogues corresponding to a given task.

        :return dialogue_ids:
            A list of two ids which correspond to the id for each conversation
        """
        return task['pairing_dict']['dialogue_ids']

    def _poll_task_queue(
        self, worker_id: str, task_data: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Poll task queue for tasks for a worker.

        :param worker_id:
            id for worker

        :param task_data:
            list of potential tasks already for worker

        :return task_data:
            a list of tasks for a worker to complete
        """
        worker_data = self._get_worker_data(worker_id)
        num_attempts = 0
        while (not self.task_queue.empty()) and num_attempts < self.task_queue.qsize():
            try:
                next_task = self.task_queue.get()
            except queue.Empty:
                break
            num_attempts += 1

            pair_id = next_task['pair_id']
            dialogue_ids = self._get_dialogue_ids(next_task)

            # make sure worker has not seen these conversations before
            if pair_id not in worker_data['tasks_completed'] and all(
                d_id not in worker_data['conversations_seen'] for d_id in dialogue_ids
            ):
                # track tasks and conversations seen
                worker_data['tasks_completed'].append(pair_id)
                worker_data['conversations_seen'].extend(dialogue_ids)
                task_data.append(next_task)
                if len(task_data) == self.opt['subtasks_per_hit']:
                    return task_data
            else:
                self.task_queue.put(next_task)

        return task_data

    def _top_up_task_data(
        self, worker_id: str, task_data: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Top up worker task data.

        This function is called if ``self.task_queue`` is exhausted but
        task_data for the worker is less than the `tasks_per_hit`.

        Make sure that all added tasks have not been seen by the worker.

        :param worker_id:
            id for worker

        :param task_data:
            list of potential tasks already for worker

        :return task_data:
            a list of tasks for a worker to complete
        """
        worker_data = self._get_worker_data(worker_id)
        tasks_still_needed = self.opt['subtasks_per_hit'] - len(task_data)
        tasks_remaining = [
            t_id
            for t_id in range(len(self.desired_tasks))
            if t_id not in worker_data['tasks_completed']
        ]
        # get any pairings with conversations this worker has not seen to fill this hit
        additional_tasks = [
            t
            for t in tasks_remaining
            if all(
                d_id not in worker_data['conversations_seen']
                for d_id in self._get_dialogue_ids(self.desired_tasks[t])
            )
        ]
        if tasks_still_needed < len(additional_tasks):
            additional_tasks = random.sample(additional_tasks, tasks_still_needed)
        worker_data['tasks_completed'].extend(additional_tasks)

        for t in additional_tasks:
            worker_data['conversations_seen'].extend(
                self._get_dialogue_ids(self.desired_tasks[t])
            )
            task_data.append(self.desired_tasks[t])

        return task_data

    def get_new_task_data(self, worker_id: str) -> List[Dict[str, Any]]:
        """
        Get next task for worker.

        Returns the next onboarding task if worker hasn't finished them all,
        Otherwise finds a task from the queue they haven't seen

        If they've seen everything in the queue, spin up an
        extra task (one that was in the queue and is now saturated)

        :param worker_id:
            worker id

        :return task_data:
            A list of tasks for the worker to complete
        """
        tasks_per_hit = self.opt['subtasks_per_hit']
        # first add onboarding tasks
        task_data = self.get_onboarding_tasks(worker_id)
        if len(task_data) == tasks_per_hit:
            return task_data

        # poll the task queue for more tasks
        task_data = self._poll_task_queue(worker_id, task_data)
        if len(task_data) == tasks_per_hit:
            return task_data

        # top up the task_data if we don't hit the desired tasks_per_hit
        task_data = self._top_up_task_data(worker_id, task_data)
        return task_data

    def requeue_task_data(self, worker_id: str, task_data: List[Dict[str, Any]]):
        """
        Return task to task_queue.

        If the task is an onboarding task, indicate that the worker has
        another onboarding task to do.

        :param worker_id:
            worker id of worker who is returning task

        :param task_data:
            list of unfinished tasks to return to the queue.
        """
        worker_data = self._get_worker_data(worker_id)
        for subtask_data in task_data:
            if subtask_data['task_specs'].get('is_onboarding', False):
                worker_data['onboarding_todo'].append(subtask_data['pair_id'])
            else:
                self.task_queue.put(subtask_data)
                try:
                    worker_data['tasks_completed'].remove(subtask_data['pair_id'])
                    for d_id in self._get_dialogue_ids(subtask_data):
                        worker_data['conversations_seen'].remove(d_id)
                except ValueError:
                    # Task may have shown up in worker's task queue twice
                    # due to some unfortunate race condition
                    warn_once(f'could not remove task from worker {worker_id} history')

    def get_onboarding_tasks(self, worker_id: str) -> List[Dict[str, Any]]:
        """
        Get next onboarding task for given worker.

        :param worker_id:
            worker id

        :return:
            A list of onboarding tasks for the worker
        """
        if len(self.onboarding_tasks) == 0:
            return []

        worker_data = self._get_worker_data(worker_id)
        onboarding_todo = worker_data['onboarding_todo']
        if not onboarding_todo:
            # worker has completed all required onboarding tasks
            return []
        # get onboarding tasks for workers needing them
        num_tasks_to_return = min(len(onboarding_todo), self.opt['subtasks_per_hit'])
        onboarding_tasks_chosen = onboarding_todo[:num_tasks_to_return]
        worker_data['onboarding_todo'] = onboarding_todo[num_tasks_to_return:]
        return [self.onboarding_tasks[t_id] for t_id in onboarding_tasks_chosen]

    def check_and_update_worker_approval(
        self, worker_id: str, save_data: Dict[str, Any]
    ):
        """
        Soft block workers who fail onboarding tasks, keep track of their status.

        :param worker_id:
            worker id

        :param save_data:
            data from the worker's completed tasks
        """
        all_task_data = save_data['worker_data'][worker_id]['task_data']
        response_data = save_data['worker_data'][worker_id]['response']['task_data']
        num_onboarding_tasks = 0
        num_correct = 0

        for i in range(len(all_task_data)):
            is_onboarding = all_task_data[i]['pairing_dict'].get('is_onboarding', False)
            if not is_onboarding:
                # not an onboarding task, no need to check correctness
                continue
            worker_response = response_data[i]['speakerChoice']
            expected_response = all_task_data[i]['pairing_dict']['correct_answer']
            num_onboarding_tasks += 1
            if worker_response == expected_response:
                # count correct answers
                num_correct += 1
        if num_onboarding_tasks == 0:
            # no onboarding tasks found
            if worker_id in self.failed_onboard:
                # worker already failed onboarding, add pairings back to queue
                self.requeue_task_data(worker_id, all_task_data)
            return
        if (num_correct / num_onboarding_tasks) >= self.opt['onboarding_threshold']:
            # worker passed onboarding
            return
        # worker failed onboarding, soft block and record
        self.manager.soft_block_worker(worker_id)
        self.failed_onboard.add(worker_id)

    def softblock_workers(self):
        """
        Softblock workers if necessary.
        """
        if not self.opt['is_sandbox'] and self.opt['softblock_list_path'] is not None:
            softblock_list = set()
            with open(self.opt['softblock_list_path']) as f:
                for line in f:
                    softblock_list.add(line.strip())
            print(f'Will softblock {len(softblock_list):d} workers.')
            for w in softblock_list:
                try:
                    print('Soft Blocking {}\n'.format(w))
                    self.manager.soft_block_worker(w)
                except Exception as e:
                    print(f'Did not soft block worker {w}: {e}')
                time.sleep(0.1)

    def run(self):
        self.manager.setup_server(
            task_directory_path=os.path.dirname(os.path.abspath(__file__))
        )
        self.manager.set_onboard_function(onboard_function=None)
        task_group_id: str = None

        try:
            # Initialize run information
            self.manager.start_new_run()

            task_group_id = self.manager.task_group_id
            self.set_block_qual(task_group_id)
            self.manager.ready_to_accept_workers()
            self.manager.create_hits()

            def check_worker_eligibility(worker):
                return True

            def assign_worker_roles(workers):
                workers[0].id = AGENT_DISPLAY_NAME

            def run_conversation(mturk_manager, opt, workers):
                task_data = self.get_new_task_data(workers[0].worker_id)
                world = StaticMTurkTaskWorld(
                    opt, mturk_agent=workers[0], task_data=task_data
                )
                while not world.episode_done():
                    world.parley()

                world.shutdown()

                save_data = world.prep_save_data(workers)

                if not world.did_complete():
                    self.requeue_task_data(workers[0].worker_id, task_data)
                else:
                    if opt['block_on_onboarding_fail']:
                        # check whether workers failed onboarding
                        self.check_and_update_worker_approval(
                            workers[0].worker_id, save_data
                        )
                return save_data

            # Soft-block all chosen workers
            self.softblock_workers()
            print("This run id: {}".format(task_group_id))

            # Begin the task, allowing mturk_manager to start running the task
            # world on any workers who connect
            self.manager.start_task(
                eligibility_function=check_worker_eligibility,
                assign_role_function=assign_worker_roles,
                task_function=run_conversation,
            )
        finally:
            self.manager.expire_all_unassigned_hits()
            self.manager.shutdown()

        return task_group_id


if __name__ == '__main__':
    args = add_args(from_argv=True)
    runner = AcuteEvaluator(args)
    runner.run()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import json
import os

from parlai.utils.io import PathManager
from parlai.core.teachers import DialogTeacher, MultiTaskTeacher
from parlai.tasks.multinli.agents import (
    convert_to_dialogData,
    BICLASS_LABELS,
    MULTINLI_LABELS,
)

from .build import build

ANLI = 'ANLI'
ANLI_PREFIX = 'anli_'
ANLI_VERSION = 'v0.1'
ANLI_LABEL_DICT = {'e': 'entailment', 'c': 'contradiction', 'n': 'neutral'}
ANLI_PREMISE_KEY = 'context'
ANLI_HYPO_KEY = 'hypothesis'
ANLI_ANSWER_KEY = 'label'
ANLI_ROUNDS = ['R1', 'R2', 'R3']


def _path(opt, anli_round):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        suffix = 'train'
    elif dt == 'valid':
        suffix = 'dev'
    elif dt == 'test':
        suffix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(
        opt['datapath'], ANLI, ANLI_PREFIX + ANLI_VERSION, anli_round, suffix + '.jsonl'
    )

    return data_path


class RoundBaseTeacher(DialogTeacher):
    """
    Base class for teachers in all 3 rounds in ANLI tasks.

    ``RoundBaseTeacher`` derives anli_round (the round index of ANLI task which consists of 3 rounds NLI tasks with
    increasing difficulty. (See https://arxiv.org/abs/1910.14599 for more information)  from ``opt['task']``.
    ``anli_round`` is used to set the correct path to the downloaded data file for thaue specified round.

    ``RoundBaseTeacher`` also parses the requested dialog format(text format w/ or w/o special tokens 'Hypothesis',
    'Premise') and number of classes (3 classes entailment, neutral, contradiction or 2 classes contradiction, not
    contradiction) desired for particular training purposes.
    """

    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('RoundBase Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        )
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        self.anli_round = (
            opt['task'].split(':')[1] if len(opt['task'].split(':')) > 1 else None
        )
        self.anli_round = self.anli_round.upper()
        if self.anli_round not in ANLI_ROUNDS:
            raise KeyError(f"Undefined task round: {self.anli_round}.")

        data_path = _path(opt, self.anli_round)
        opt['datafile'] = data_path
        self.dialog_format = opt.get('dialog_format', False)
        self.binary_classes = opt.get('binary_classes', False)
        self.id = opt['task'].upper()
        super().__init__(opt, shared)

    def label_candidates(self):
        if self.binary_classes:
            return BICLASS_LABELS
        return MULTINLI_LABELS

    def setup_data(self, path):
        print('loading: ' + path)
        with PathManager.open(path, 'r') as data_file:
            for pair_line in data_file:
                pair = json.loads(pair_line)
                if pair[ANLI_ANSWER_KEY] == '-':
                    continue

                label_raw = pair[ANLI_ANSWER_KEY]
                if label_raw in ANLI_LABEL_DICT:
                    label_raw = ANLI_LABEL_DICT[label_raw]

                question, answers, clas = convert_to_dialogData(
                    premise_raw=pair[ANLI_PREMISE_KEY],
                    hypo_raw=pair[ANLI_HYPO_KEY],
                    answer_raw=label_raw,
                    dialog_format=self.dialog_format,
                    binary_classes=self.binary_classes,
                )

                yield (question, answers, None, clas), True


class R1Teacher(RoundBaseTeacher):
    pass


class R2Teacher(RoundBaseTeacher):
    pass


class R3Teacher(RoundBaseTeacher):
    pass


class DefaultTeacher(MultiTaskTeacher):
    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('ANLI Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        ),
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )

    def __init__(self, opt, shared=None):
        anli_tasks = ['anli:r1', 'anli:r2', 'anli:r3']
        opt = copy.deepcopy(opt)
        opt['task'] = ','.join(anli_tasks)
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Any

task_config: Dict[str, Any] = {}


task_config['frontend_version'] = 1

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Chat with two other people'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config['hit_description'] = 'Chat with two other people.'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,dialog'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''\'\'\'
In this task, you are going to chat with two other people.<br><br>

If you are ready, please click "Accept HIT" to start this task.
\'\'\''''



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import os
from unittest import mock
from parlai.mturk.core.dev.worker_manager import WorkerManager, WorkerState
from parlai.mturk.core.dev.agents import MTurkAgent, AssignState
from parlai.mturk.core.dev.mturk_manager import MTurkManager
from parlai.core.params import ParlaiParser

import parlai.mturk.core.dev.worker_manager as WorkerManagerFile
import parlai.mturk.core.dev.data_model as data_model

parent_dir = os.path.dirname(os.path.abspath(__file__))
WorkerManagerFile.DISCONNECT_FILE_NAME = 'disconnect-test.pickle'
WorkerManagerFile.MAX_DISCONNECTS = 1
WorkerManagerFile.parent_dir = os.path.dirname(os.path.abspath(__file__))

TEST_WORKER_ID_1 = 'TEST_WORKER_ID_1'
TEST_WORKER_ID_2 = 'TEST_WORKER_ID_2'
TEST_WORKER_ID_3 = 'TEST_WORKER_ID_3'
TEST_ASSIGNMENT_ID_1 = 'TEST_ASSIGNMENT_ID_1'
TEST_ASSIGNMENT_ID_2 = 'TEST_ASSIGNMENT_ID_2'
TEST_ASSIGNMENT_ID_3 = 'TEST_ASSIGNMENT_ID_3'
TEST_HIT_ID_1 = 'TEST_HIT_ID_1'
TEST_HIT_ID_2 = 'TEST_HIT_ID_2'
TEST_HIT_ID_3 = 'TEST_HIT_ID_3'
FAKE_ID = 'BOGUS'


class TestWorkerState(unittest.TestCase):
    """
    Various unit tests for the WorkerState class.
    """

    def setUp(self):
        self.work_state_1 = WorkerState(TEST_WORKER_ID_1, 10)
        self.work_state_2 = WorkerState(TEST_WORKER_ID_2)
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args()
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        mturk_agent_ids = ['mturk_agent_1']
        self.mturk_manager = MTurkManager(opt=self.opt, mturk_agent_ids=mturk_agent_ids)
        self.mturk_manager.send_message = mock.MagicMock()
        self.mturk_manager.send_state_change = mock.MagicMock()
        self.mturk_manager.send_command = mock.MagicMock()
        self.worker_manager = WorkerManager(self.mturk_manager, self.opt)

    def tearDown(self):
        self.mturk_manager.shutdown()

    def test_worker_state_init(self):
        """
        Test proper initialization of worker states.
        """
        self.assertEqual(self.work_state_1.worker_id, TEST_WORKER_ID_1)
        self.assertEqual(self.work_state_2.worker_id, TEST_WORKER_ID_2)
        self.assertEqual(self.work_state_1.disconnects, 10)
        self.assertEqual(self.work_state_2.disconnects, 0)

    def test_worker_state_agent_management(self):
        """
        Test public state management methods of worker_state.
        """
        agent_1 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_1,
            TEST_ASSIGNMENT_ID_1,
            TEST_WORKER_ID_1,
        )
        agent_2 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_2,
            TEST_ASSIGNMENT_ID_2,
            TEST_WORKER_ID_1,
        )
        agent_3 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_3,
            TEST_ASSIGNMENT_ID_3,
            TEST_WORKER_ID_3,
        )

        self.assertEqual(self.work_state_1.active_conversation_count(), 0)
        self.work_state_1.add_agent(agent_1)
        self.assertEqual(self.work_state_1.active_conversation_count(), 1)
        self.work_state_1.add_agent(agent_2)
        self.assertEqual(self.work_state_1.active_conversation_count(), 2)

        with self.assertRaises(AssertionError):
            self.work_state_1.add_agent(agent_3)

        self.assertEqual(self.work_state_1.active_conversation_count(), 2)
        self.assertEqual(self.work_state_1.completed_assignments(), 0)

        self.assertTrue(self.work_state_1.has_assignment(agent_1.assignment_id))
        self.assertTrue(self.work_state_1.has_assignment(agent_2.assignment_id))
        self.assertFalse(self.work_state_1.has_assignment(agent_3.assignment_id))
        self.assertEqual(
            agent_1, self.work_state_1.get_agent_for_assignment(agent_1.assignment_id)
        )
        self.assertEqual(
            agent_2, self.work_state_1.get_agent_for_assignment(agent_2.assignment_id)
        )
        self.assertIsNone(
            self.work_state_1.get_agent_for_assignment(agent_3.assignment_id)
        )

        agent_1.set_status(AssignState.STATUS_DONE)
        self.assertEqual(self.work_state_1.active_conversation_count(), 1)
        self.assertEqual(self.work_state_1.completed_assignments(), 1)
        agent_2.set_status(AssignState.STATUS_DISCONNECT)
        self.assertEqual(self.work_state_1.active_conversation_count(), 0)
        self.assertEqual(self.work_state_1.completed_assignments(), 1)

    def test_manager_alive_makes_state(self):
        test_worker = self.worker_manager.worker_alive(TEST_WORKER_ID_1)
        self.assertIsInstance(test_worker, WorkerState)
        self.assertEqual(test_worker.worker_id, TEST_WORKER_ID_1)
        self.assertNotEqual(test_worker, self.work_state_1)


class TestWorkerManager(unittest.TestCase):
    """
    Various unit tests for the WorkerManager class.
    """

    def setUp(self):
        disconnect_path = os.path.join(parent_dir, 'disconnect-test.pickle')
        if os.path.exists(disconnect_path):
            os.remove(disconnect_path)

        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args()
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        mturk_agent_ids = ['mturk_agent_1']
        self.mturk_manager = MTurkManager(
            opt=self.opt.copy(), mturk_agent_ids=mturk_agent_ids
        )
        self.worker_manager = self.mturk_manager.worker_manager
        self.mturk_manager.send_message = mock.MagicMock()
        self.mturk_manager.send_state_change = mock.MagicMock()
        self.mturk_manager.send_command = mock.MagicMock()

        self.worker_state_1 = self.worker_manager.worker_alive(TEST_WORKER_ID_1)
        self.worker_state_2 = self.worker_manager.worker_alive(TEST_WORKER_ID_2)
        self.worker_state_3 = self.worker_manager.worker_alive(TEST_WORKER_ID_3)

    def tearDown(self):
        self.mturk_manager.shutdown()

        disconnect_path = os.path.join(parent_dir, 'disconnect-test.pickle')
        if os.path.exists(disconnect_path):
            os.remove(disconnect_path)

    def test_private_create_agent(self):
        """
        Check create agent method used internally in worker_manager.
        """
        test_agent = self.worker_manager._create_agent(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_1, TEST_WORKER_ID_1
        )
        self.assertIsInstance(test_agent, MTurkAgent)
        self.assertEqual(test_agent.worker_id, TEST_WORKER_ID_1)
        self.assertEqual(test_agent.hit_id, TEST_HIT_ID_1)
        self.assertEqual(test_agent.assignment_id, TEST_ASSIGNMENT_ID_1)

    def test_agent_task_management(self):
        """
        Ensure agents and tasks have proper bookkeeping.
        """
        self.worker_manager.assign_task_to_worker(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_1, TEST_WORKER_ID_1
        )
        self.worker_manager.assign_task_to_worker(
            TEST_HIT_ID_2, TEST_ASSIGNMENT_ID_2, TEST_WORKER_ID_2
        )
        self.worker_manager.assign_task_to_worker(
            TEST_HIT_ID_3, TEST_ASSIGNMENT_ID_3, TEST_WORKER_ID_1
        )
        self.assertTrue(self.worker_state_1.has_assignment(TEST_ASSIGNMENT_ID_1))
        self.assertTrue(self.worker_state_1.has_assignment(TEST_ASSIGNMENT_ID_3))
        self.assertTrue(self.worker_state_2.has_assignment(TEST_ASSIGNMENT_ID_2))

        assign_agent = self.worker_manager.get_agent_for_assignment(
            TEST_ASSIGNMENT_ID_1
        )
        self.assertEqual(assign_agent.worker_id, TEST_WORKER_ID_1)
        self.assertEqual(assign_agent.hit_id, TEST_HIT_ID_1)
        self.assertEqual(assign_agent.assignment_id, TEST_ASSIGNMENT_ID_1)

        no_such_agent = self.worker_manager.get_agent_for_assignment(FAKE_ID)
        self.assertIsNone(no_such_agent)

        # Ensure all agents are being maintained
        checked_count = 0
        filtered_count = 0

        def check_is_worker_1(agent):
            nonlocal checked_count  # noqa E999 python 3 only
            checked_count += 1
            self.assertEqual(agent.worker_id, TEST_WORKER_ID_1)

        def is_worker_1(agent):
            nonlocal filtered_count
            filtered_count += 1
            return agent.worker_id == TEST_WORKER_ID_1

        self.worker_manager.map_over_agents(check_is_worker_1, is_worker_1)
        self.assertEqual(checked_count, 2)
        self.assertEqual(filtered_count, 3)

        # Ensuring _get_worker is accurate
        self.assertEqual(
            self.worker_manager._get_worker(TEST_WORKER_ID_1), self.worker_state_1
        )
        self.assertEqual(
            self.worker_manager._get_worker(TEST_WORKER_ID_2), self.worker_state_2
        )
        self.assertEqual(
            self.worker_manager._get_worker(TEST_WORKER_ID_3), self.worker_state_3
        )
        self.assertIsNone(self.worker_manager._get_worker(FAKE_ID))

        # Ensuring _get_agent is accurate
        self.assertEqual(
            self.worker_manager._get_agent(TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1),
            self.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_1),
        )
        self.assertNotEqual(
            self.worker_manager._get_agent(TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_2),
            self.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_2),
        )

    def test_shutdown(self):
        """
        Ensure shutdown clears required resources.
        """
        self.worker_manager.save_disconnects = mock.MagicMock()
        self.worker_manager.un_time_block_workers = mock.MagicMock()
        self.worker_manager.shutdown()
        self.assertEqual(
            len(self.worker_manager.save_disconnects.mock_calls),
            1,
            'save_disconnects must be called in worker manager shutdown',
        )
        self.assertEqual(
            len(self.worker_manager.un_time_block_workers.mock_calls),
            1,
            'un_time_block_workers must be called in worker manager shutdown',
        )

    def test_time_blocks(self):
        """
        Check to see if time blocking and clearing works.
        """
        self.mturk_manager.soft_block_worker = mock.MagicMock()
        self.mturk_manager.un_soft_block_worker = mock.MagicMock()

        # No workers blocked, none should be unblocked
        self.worker_manager.un_time_block_workers()
        self.assertEqual(len(self.mturk_manager.un_soft_block_worker.mock_calls), 0)

        # Block some workers, ensure state change and correct calls
        self.assertEqual(len(self.worker_manager.time_blocked_workers), 0)
        self.worker_manager.time_block_worker(TEST_WORKER_ID_1)
        self.mturk_manager.soft_block_worker.assert_called_with(
            TEST_WORKER_ID_1, 'max_time_qual'
        )
        self.assertEqual(len(self.worker_manager.time_blocked_workers), 1)
        self.worker_manager.time_block_worker(TEST_WORKER_ID_2)
        self.mturk_manager.soft_block_worker.assert_called_with(
            TEST_WORKER_ID_2, 'max_time_qual'
        )
        self.assertEqual(len(self.worker_manager.time_blocked_workers), 2)
        self.assertEqual(len(self.mturk_manager.soft_block_worker.mock_calls), 2)

        # Unblock a worker passed in as a keyword arg, ensure state remains
        self.worker_manager.un_time_block_workers([TEST_WORKER_ID_3])
        self.mturk_manager.un_soft_block_worker.assert_called_with(
            TEST_WORKER_ID_3, 'max_time_qual'
        )
        self.assertEqual(len(self.worker_manager.time_blocked_workers), 2)

        # Unblock blocked workers, ensure proper calls and state change
        self.worker_manager.un_time_block_workers()
        self.assertEqual(len(self.worker_manager.time_blocked_workers), 0)
        self.mturk_manager.un_soft_block_worker.assert_any_call(
            TEST_WORKER_ID_1, 'max_time_qual'
        )
        self.mturk_manager.un_soft_block_worker.assert_any_call(
            TEST_WORKER_ID_2, 'max_time_qual'
        )

    def test_disconnect_management(self):
        self.worker_manager.load_disconnects()
        self.worker_manager.is_sandbox = False
        self.mturk_manager.block_worker = mock.MagicMock()
        self.mturk_manager.soft_block_worker = mock.MagicMock()

        self.assertEqual(len(self.worker_manager.disconnects), 0)
        # Make one worker disconnect twice
        self.worker_manager.handle_bad_disconnect(TEST_WORKER_ID_1)
        self.assertEqual(len(self.worker_manager.disconnects), 1)
        self.mturk_manager.block_worker.assert_not_called()
        self.mturk_manager.soft_block_worker.assert_not_called()
        self.worker_manager.handle_bad_disconnect(TEST_WORKER_ID_1)
        self.assertEqual(len(self.worker_manager.disconnects), 2)
        self.mturk_manager.block_worker.assert_not_called()
        self.mturk_manager.soft_block_worker.assert_not_called()

        # Ensure both disconnects recorded
        self.assertEqual(
            self.worker_manager.mturk_workers[TEST_WORKER_ID_1].disconnects, 2
        )

        # Make second worker disconnect
        self.worker_manager.handle_bad_disconnect(TEST_WORKER_ID_2)
        self.assertEqual(len(self.worker_manager.disconnects), 3)
        self.mturk_manager.block_worker.assert_not_called()
        self.mturk_manager.soft_block_worker.assert_not_called()

        # Make us soft block workers on disconnect
        self.worker_manager.opt['disconnect_qualification'] = 'test'
        self.worker_manager.handle_bad_disconnect(TEST_WORKER_ID_1)
        self.mturk_manager.block_worker.assert_not_called()
        self.mturk_manager.soft_block_worker.assert_called_with(
            TEST_WORKER_ID_1, 'disconnect_qualification'
        )
        self.mturk_manager.soft_block_worker.reset_mock()

        # Make us now block workers on disconnect
        self.worker_manager.opt['hard_block'] = True
        self.worker_manager.handle_bad_disconnect(TEST_WORKER_ID_2)
        self.mturk_manager.block_worker.assert_called_once()
        self.mturk_manager.soft_block_worker.assert_not_called()

        # Ensure we can save and reload disconnects
        self.worker_manager.save_disconnects()

        # Make a new worker manager
        worker_manager2 = WorkerManager(self.mturk_manager, self.opt)
        self.assertEqual(len(worker_manager2.disconnects), 5)
        self.assertEqual(worker_manager2.mturk_workers[TEST_WORKER_ID_1].disconnects, 3)
        self.assertEqual(worker_manager2.mturk_workers[TEST_WORKER_ID_2].disconnects, 2)
        worker_manager2.shutdown()

    def test_conversation_management(self):
        """
        Tests handling conversation state, moving agents to the correct conversations,
        and disconnecting one worker in an active convo.
        """
        self.worker_manager.assign_task_to_worker(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_1, TEST_WORKER_ID_1
        )
        self.worker_manager.assign_task_to_worker(
            TEST_HIT_ID_2, TEST_ASSIGNMENT_ID_2, TEST_WORKER_ID_2
        )

        good_agent = self.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_1)
        bad_agent = self.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_2)

        def fake_command_send(worker_id, assignment_id, data, ack_func):
            pkt = mock.MagicMock()
            pkt.sender_id = worker_id
            pkt.assignment_id = assignment_id
            self.assertEqual(data['text'], data_model.COMMAND_CHANGE_CONVERSATION)
            ack_func(pkt)

        self.mturk_manager.send_command = fake_command_send
        good_agent.set_status(
            AssignState.STATUS_IN_TASK, conversation_id='t1', agent_id='good'
        )
        bad_agent.set_status(
            AssignState.STATUS_IN_TASK, conversation_id='t1', agent_id='bad'
        )
        self.assertEqual(good_agent.id, 'good')
        self.assertEqual(bad_agent.id, 'bad')
        self.assertEqual(good_agent.conversation_id, 't1')
        self.assertEqual(bad_agent.conversation_id, 't1')
        self.assertIn('t1', self.worker_manager.conv_to_agent)
        self.assertEqual(len(self.worker_manager.conv_to_agent['t1']), 2)
        self.worker_manager.handle_bad_disconnect = mock.MagicMock()

        checked_worker = False

        def partner_callback(agent):
            nonlocal checked_worker
            checked_worker = True
            self.assertEqual(agent.worker_id, good_agent.worker_id)

        self.worker_manager.handle_agent_disconnect(
            bad_agent.worker_id, bad_agent.assignment_id, partner_callback
        )
        self.assertTrue(checked_worker)
        self.worker_manager.handle_bad_disconnect.assert_called_once_with(
            bad_agent.worker_id
        )
        self.assertEqual(bad_agent.get_status(), AssignState.STATUS_DISCONNECT)


if __name__ == '__main__':
    unittest.main(buffer=True)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkOnboardWorld
from parlai.mturk.core.agents import TIMEOUT_MESSAGE
from parlai.core.worlds import validate, MultiAgentDialogWorld
from parlai.utils.safety import OffensiveStringMatcher
from joblib import Parallel, delayed
from task_config import task_config as config
from extract_and_save_personas import main as main_extract
from constants import (
    MAX_DOC_LEN,
    WIZARD,
    APPRENTICE,
    ONBOARD_MSG,
    APPRENTICE_START_MSG,
    WIZARD_START_MSG,
    TIMEOUT_MSG,
    EXCEED_MIN_TURNS_MSG,
    UNEXPECTED_DISCONNECTION_MSG,
    CHAT_ENDED_MSG,
    STOPWORDS,
    EVAL_WIZARD_MSG,
    PARTNER_RETRIEVED_PASSAGES_INST_MSG,
    WAITING_MSG,
    PICK_TOPIC_MSG,
    AFTER_PICK_TOPIC_MSG,
    AFTER_PICK_TOPIC_WIZARD_MSG,
    AFTER_PARTNER_PICK_TOPIC_WIZARD_MSG,
)
import numpy as np
import time
import os
import pickle
import random
import copy
from urllib.parse import unquote


def split_tokenize(text):
    """
    Splits tokens based on whitespace after adding whitespace around punctuation.
    """
    return (
        text.replace('.', ' . ')
        .replace('. . .', '...')
        .replace(',', ' , ')
        .replace(';', ' ; ')
        .replace(':', ' : ')
        .replace('!', ' ! ')
        .replace('?', ' ? ')
        .replace('(', ' ( ')
        .replace(')', ' ) ')
        .split()
    )


class PersonasGenerator(object):
    def __init__(self, opt):
        self.personas_idx_stack_path = os.path.join(
            os.getcwd(), './personas_idx_stack.pkl'
        )

        self.personas_path = '{}/data/personas-{}'.format(
            os.getcwd(), opt['persona_type'] + 'Original'
        )
        self.topic_for_personas_path = '{}/personas_with_wiki_links.txt'.format(
            os.getcwd()
        )
        if not os.path.exists(self.personas_path):
            opt['personas_path'] = self.personas_path
            main_extract(opt)
        self.personas_name_list = []

        for f_name in os.listdir(self.personas_path):
            if f_name.endswith('.pkl'):
                self.personas_name_list.append(f_name)

        if os.path.exists(self.personas_idx_stack_path):
            with open(self.personas_idx_stack_path, 'rb') as handle:
                self.idx_stack = pickle.load(handle)
        else:
            self.idx_stack = []
            self.add_idx_stack()
            self.save_idx_stack()

        self.load_topics_for_personas()

    def add_idx_stack(self):
        stack = [i for i in range(len(self.personas_name_list))]
        random.seed()
        random.shuffle(stack)
        self.idx_stack = stack + self.idx_stack

    def pop_persona(self):
        if len(self.idx_stack) == 0:
            self.add_idx_stack()
        idx = self.idx_stack.pop()
        data = np.load(
            os.path.join(self.personas_path, self.personas_name_list[idx]),
            allow_pickle=True,
        )
        return (idx, data)

    def push_persona(self, idx):
        self.idx_stack.append(idx)

    def save_idx_stack(self):
        with open(self.personas_idx_stack_path, 'wb') as handle:
            pickle.dump(self.idx_stack, handle)

    def load_topics_for_personas(self):
        self.persona_to_topics = {}
        with open(self.topic_for_personas_path) as f:
            text = f.read()
            personas = text.split('\n\n')
            for persona in personas:
                persona = persona.split('\n')
                prev_p = persona[0]
                for i in range(1, len(persona)):
                    p_i = persona[i]
                    if 'https' in p_i:
                        topic = unquote(p_i[p_i.rfind('/') + 1 :]).replace('_', ' ')
                        if prev_p in self.persona_to_topics:
                            self.persona_to_topics[prev_p].append(topic)
                        else:
                            self.persona_to_topics[prev_p] = [topic]
                    else:
                        prev_p = p_i

    def get_topics(self, persona_sent):
        return self.persona_to_topics.get(persona_sent, [])


class RoleOnboardWorld(MTurkOnboardWorld):
    """
    A world that provides the appropriate instructions during onboarding.
    """

    def __init__(self, opt, mturk_agent, role):
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.max_onboard_time = opt['max_onboard_time']
        self.role = role
        super().__init__(opt, mturk_agent)

    def parley(self):
        onboard_msg = {'id': 'SYSTEM'}

        onboard_msg['show_persona'] = False
        onboard_msg['text'] = ONBOARD_MSG
        if self.role == WIZARD:
            onboard_msg['role_task_description'] = config['wizard_onboarding']
        else:
            onboard_msg['role_task_description'] = config['apprentice_onboarding']
        self.mturk_agent.observe(onboard_msg)

        act = self.mturk_agent.act(timeout=self.max_onboard_time)
        # timeout
        if act['episode_done'] or (('text' in act and act['text'] == TIMEOUT_MESSAGE)):
            self.episodeDone = True
            return

        if 'text' not in act:
            control_msg = {'id': 'SYSTEM', 'text': WAITING_MSG}
            self.mturk_agent.observe(validate(control_msg))
            self.episodeDone = True


class MTurkWizardOfWikipediaWorld(MultiAgentDialogWorld):
    """
    World where two agents have a dialogue; one chats freely, perhaps based on a
    persona, while the other is the 'wizard', who bases his/her responses on documents
    (i.e. sentences) retrieved based on what the other agent says.
    """

    def __init__(
        self,
        opt,
        agents=None,
        shared=None,
        world_tag='NONE',
        ir_agent=None,
        task='',
        wiki_title_to_passage=None,
    ):
        self.turn_idx = 0
        self.min_turns = opt['min_turns']
        self.max_turns = opt['max_turns']
        self.num_turns = np.random.randint(self.min_turns, self.max_turns) + 1
        self.dialog = []
        self.wizard_eval = 0
        self.task_type = 'sandbox' if opt['is_sandbox'] else 'live'
        self.chat_done = False
        self.world_tag = world_tag
        self.max_resp_time = opt['max_resp_time']  # in secs
        self.num_passages_to_retrieve = opt['num_passages_retrieved']
        super().__init__(opt, agents, shared)
        self.agents = sorted(agents, key=lambda x: x.id, reverse=random.random() <= 0.5)
        #  Personas and retriever
        self.persona_generator = self.agents[0].persona_generator
        self.relevant_topics = []
        while not self.relevant_topics:
            self.persona_to_topics = {}
            self.persona_idx, persona_data = self.persona_generator.pop_persona()
            for p in persona_data:
                if p[0] == ' ':
                    p = p[1:]
                if p not in self.persona_to_topics:
                    self.persona_to_topics[p] = []
                    topics = set(self.persona_generator.get_topics(p))
                    for t in topics:
                        self.relevant_topics.append(t + ' ({})'.format(p))
                        self.persona_to_topics[p].append(t)

        self.ir_agent = ir_agent
        self.setup_tokenizer(opt)
        self.chosen_topic = ''
        self.chosen_topic_passage = {}
        self.OLD = OffensiveStringMatcher()
        # Load the title to passage dictionary
        self.wiki_title_to_passage = wiki_title_to_passage

    def episode_done(self):
        return self.chat_done

    def setup_tokenizer(self, opt):
        try:
            import nltk
        except ImportError:
            raise ImportError('Please install nltk (e.g. pip install nltk).')
        # nltk-specific setup
        st_path = 'tokenizers/punkt/{0}.pickle'.format(opt['dict_language'])
        try:
            self.sent_tok = nltk.data.load(st_path)
        except LookupError:
            nltk.download('punkt')
            self.sent_tok = nltk.data.load(st_path)

    def sufficient_overlap(self, text, sent_dict):
        text_list = [w[:4] for w in split_tokenize(text.lower()) if w not in STOPWORDS]
        for _, sentence in sent_dict.items():
            sentence_list = [
                w[:4] for w in split_tokenize(sentence.lower()) if w not in STOPWORDS
            ]
            if len(set(text_list).intersection(set(sentence_list))) >= self.opt.get(
                'word_overlap_threshold', 2
            ):
                return True
        return False

    def parley(self):
        """
        Each agent acts; when the APPRENTICE says something, the WIZARD is given
        retrieved documents based on the text response.
        """
        self.turn_idx += 1

        # Initial Message Value
        control_msg = {'episode_done': False}
        control_msg['id'] = 'SYSTEM'

        print(self.world_tag + ' is at turn {}...'.format(self.turn_idx))

        '''First Turn: We give the first agent the list of topics to choose from
        '''
        if self.turn_idx == 1:
            for idx, agent in enumerate(self.agents):
                """
                If we are giving the persona, do that :)
                """
                control_msg['text'] = self.get_instruction(
                    tag='start', agent_id=agent.id
                )
                if agent.id == WIZARD:
                    control_msg['description'] = config['wizard_onboarding']
                else:
                    control_msg['description'] = config['apprentice_onboarding']
                agent.observe(validate(control_msg))
                if idx == 0:
                    time.sleep(3)

            '''Send First Person the list of relevant topics'''
            self.agents[0].observe(
                validate(
                    {
                        'id': 'SYSTEM',
                        'text': PICK_TOPIC_MSG,
                        'relevant_topics': self.relevant_topics,
                    }
                )
            )

            topic_act = self.agents[0].act(timeout=self.max_resp_time)
            timed_out = self.check_timeout(topic_act)
            if not timed_out:
                if self.agents[0].id == APPRENTICE:
                    pick_msg = AFTER_PICK_TOPIC_MSG
                else:
                    pick_msg = AFTER_PICK_TOPIC_WIZARD_MSG
                self.agents[0].observe({'id': 'SYSTEM', 'text': pick_msg})
            self.chosen_topic = topic_act['text']

            '''Now, send the wiki page for the chosen topic to the wizard'''
            for idx, agent in enumerate(self.agents):
                if agent.id == WIZARD:
                    passage = self.wiki_title_to_passage.get(self.chosen_topic, '')
                    if passage == '':
                        break
                    split = passage.split('\n')
                    title = split[0]
                    split = self.sent_tok.tokenize(" ".join(split[1:]))
                    split[0] = split[0][1:]
                    sentences = []
                    for sent in split:
                        if len(sent) > 1:
                            sentences.append(sent)
                            if len(" ".join(sentences)) > MAX_DOC_LEN * 2:
                                break
                    msg_text = AFTER_PARTNER_PICK_TOPIC_WIZARD_MSG if idx == 1 else ""
                    control_msg['text'] = msg_text
                    control_msg['chosen_topic_passages'] = [[title, sentences]]
                    agent.observe(validate(control_msg))
                    self.chosen_topic_passage = {
                        'topic': self.chosen_topic,
                        'full_passage': passage,
                        'shown_passage': sentences,
                    }

        '''If we get to the min turns, inform turker that they can end if they
           want
        '''
        if self.turn_idx == self.num_turns + 1:
            for agent in self.agents:
                control_msg['text'] = self.get_instruction(tag='exceed_min_turns')
                control_msg['exceed_min_turns'] = True
                agent.observe(validate(control_msg))

        '''Otherwise, we proceed accordingly'''
        acts = self.acts
        for idx, agent in enumerate(self.agents):
            # Increase response time for wizard
            max_response_time = self.max_resp_time * (
                1 if agent.id == APPRENTICE else 1.5
            )
            acts[idx] = agent.act(timeout=max_response_time)
            self.check_timeout(acts[idx])

            # If chat ends
            if acts[idx]['episode_done']:
                self.chat_done = True
                for ag in self.agents:
                    if ag != agent and ag.some_agent_disconnected:
                        control_msg['text'] = UNEXPECTED_DISCONNECTION_MSG
                        ag.observe(validate(control_msg))
                        return
                if self.turn_idx > self.num_turns:
                    for ag in self.agents:
                        ag.observe(validate(acts[idx]))
                        '''Have Apprentice Agent Eval Wizard Agent'''
                        if ag.id == APPRENTICE:
                            control_msg['text'] = EVAL_WIZARD_MSG
                            control_msg['wizard_eval'] = True
                            ag.observe(validate(control_msg))
                            act = ag.act(timeout=self.max_resp_time)
                            self.check_timeout(act)
                            try:
                                w_ev = int(act['text'])
                                w_ev = max(w_ev, 1) if w_ev <= 0 else min(w_ev, 5)
                            except ValueError:  # If there is a disconnect here
                                w_ev = -1
                            self.wizard_eval = w_ev
                        control_msg['text'] = CHAT_ENDED_MSG
                        ag.observe(validate(control_msg))
                return

            '''Set up msg info dict to save in dialog'''
            msg_info = {
                'speaker': '{}_{}'.format(idx, agent.id),
                'text': acts[idx]['text'],
                'turn': self.turn_idx,
                'time': time.time(),
                'offensive': self.OLD.contains_offensive_language(acts[idx]['text']),
            }

            '''Get clicked passages and checked sentences from Wizard'''
            if 'clicked_passages' in acts[idx]:
                msg_info['clicked_passages'] = acts[idx]['clicked_passages']
                checked_sents = {}
                for k, v in acts[idx]['checked_sentences'].items():
                    if k == 'no_passages_used':
                        checked_sents[k] = v
                    else:
                        split = k.split('_')
                        person = split[0]
                        topic_idx = split[1]
                        sent_idx = split[2]
                        if person == 'partner':
                            sub_passages = [
                                p.split('\n')[0]
                                for p in self.dialog[-1]['full_passages']
                            ]
                            topic = sub_passages[int(topic_idx)]
                        elif person == 'self':
                            sub_passages = [
                                p.split('\n')[0]
                                for p in self.dialog[-2]['full_passages']
                            ]
                            topic = sub_passages[int(topic_idx)]
                        else:
                            topic = self.chosen_topic
                        cs_key = '_'.join(
                            [person, '_'.join(topic.split(' ')), sent_idx]
                        )
                        checked_sents[cs_key] = v
                msg_info['checked_sentence'] = checked_sents
                msg_info['checked_passage'] = acts[idx]['checked_passages']
                msg_info['good_message'] = self.sufficient_overlap(
                    msg_info['text'], msg_info['checked_sentence']
                )

            '''Retrieve Passages'''
            ir_passages = self.retrieve_passages(copy.deepcopy(acts[idx]))
            passages = self.format_passages(ir_passages)
            msg_info['full_passages'] = ir_passages
            msg_info['shown_passages'] = passages

            if agent.id == WIZARD:
                """
                Give Wizard the Relevant Passages.
                """
                control_msg['text'] = ''
                control_msg['self_retrieved_passages'] = passages
                agent.observe(validate(control_msg))

            self.dialog.append(msg_info)

            for other_agent in self.agents:
                if other_agent != agent:
                    other_agent.observe(validate(acts[idx]))
                    if other_agent.id == WIZARD:
                        control_msg['text'] = PARTNER_RETRIEVED_PASSAGES_INST_MSG
                        control_msg['partner_retrieved_passages'] = passages
                        other_agent.observe(validate(control_msg))

    def format_passages(self, ir_passages, max_length=MAX_DOC_LEN):
        passages = []
        if len(ir_passages) == 1:  # Didn't receive any passages
            passages.append(['No Passages Retrieved', []])
        else:
            for passage in ir_passages:
                split = passage.split('\n')
                title = split[0]
                split = self.sent_tok.tokenize(" ".join(split[1:]))
                split[0] = split[0][1:]
                sentences = []
                for sent in split:
                    if len(sent) > 1:
                        sentences.append(sent)
                        if len(" ".join(sentences)) > max_length:
                            break
                passages.append([title, sentences])
        return passages

    def retrieve_passages(self, act, num_passages=None):
        if not num_passages:
            num_passages = self.num_passages_to_retrieve
        self.ir_agent.observe(act)
        action = self.ir_agent.act()
        passages = action.get('text_candidates', [action.get('text', "")])
        return passages[: min(len(passages), num_passages)]

    def get_instruction(self, agent_id=None, tag='first'):
        if tag == 'start':
            start_msg = WIZARD_START_MSG if agent_id == WIZARD else APPRENTICE_START_MSG
            return start_msg.format(self.num_turns)
        if tag == 'timeout':
            return TIMEOUT_MSG
        if tag == 'exceed_min_turns':
            return EXCEED_MIN_TURNS_MSG.format(self.num_turns)

    def check_timeout(self, act):
        if act['text'] == '[TIMEOUT]' and act['episode_done']:
            control_msg = {'episode_done': True}
            control_msg['id'] = 'SYSTEM'
            control_msg['text'] = self.get_instruction(tag='timeout')
            for ag in self.agents:
                if ag.id != act['id']:
                    ag.observe(validate(control_msg))
            self.chat_done = True
            return True
        else:
            return False

    def reset_random(self):
        self.num_turns = np.random.randint(self.min_turns, self.max_turns) + 1

    def save_data(self):
        # save persona_idx_stack
        convo_finished = self.turn_idx >= self.num_turns + 1
        for ag in self.agents:
            if (
                ag.hit_is_abandoned
                or ag.hit_is_returned
                or ag.disconnected
                or ag.hit_is_expired
            ):
                convo_finished = False
        if not convo_finished:
            self.persona_generator.push_persona(self.persona_idx)
            print("\n**Push persona {} back to stack. **\n".format(self.persona_idx))
        self.agents[0].persona_generator.save_idx_stack()

        data_path = self.opt['data_path']
        if not os.path.exists(data_path):
            os.makedirs(data_path)
        self.convo_finished = convo_finished
        self.wizard_worker = ''
        if convo_finished:
            filename = os.path.join(
                data_path,
                '{}_{}_{}.pkl'.format(
                    time.strftime("%Y%m%d-%H%M%S"),
                    np.random.randint(0, 1000),
                    self.task_type,
                ),
            )
            self.good_wiz, self.wizard_worker = self.check_wizard_quality()
        else:
            filename = os.path.join(
                data_path,
                '{}_{}_{}_incomplete.pkl'.format(
                    time.strftime("%Y%m%d-%H%M%S"),
                    np.random.randint(0, 1000),
                    self.task_type,
                ),
            )
            self.good_wiz = True
        pickle.dump(
            {
                'persona': self.persona_to_topics,
                'relevant_topics': self.relevant_topics,
                'chosen_topic_passage': self.chosen_topic_passage,
                'dialog': self.dialog,
                'speaker_with_persona': self.agents[0].worker_id,
                'workers': [ag.worker_id for ag in self.agents],
                'n_turn': self.num_turns,
                'hit_ids': [ag.hit_id for ag in self.agents],
                'assignment_ids': [ag.assignment_id for ag in self.agents],
                'wizard_eval': self.wizard_eval,
                'chosen_topic': self.chosen_topic,
                'wizard_good': convo_finished and self.good_wiz,
                'good_wizard_worker': self.wizard_worker if self.good_wiz else '',
                'bad_wizard_worker': self.wizard_worker if not self.good_wiz else '',
            },
            open(filename, 'wb'),
        )
        print('{}: Data successfully saved at {}.'.format(self.world_tag, filename))

    def check_wizard_quality(self):
        """
        Determines whether to soft-block this turker or not Only called if the
        conversation finishes Returns True if the Wizard is good.
        """
        num_good_sents = len(
            list(
                filter(
                    lambda info: 'good_message' in info and info['good_message'],
                    self.dialog,
                )
            )
        )
        wizard_worker = [w for w in self.agents if w.id == WIZARD][0].worker_id
        data_path = self.opt['current_working_dir']
        bad_wizards = os.path.join(data_path, 'bad_wizards.txt')
        good_wizards = os.path.join(data_path, 'good_wizards.txt')
        if num_good_sents < self.opt['num_good_sentence_threshold']:
            if not self.opt['is_sandbox']:
                with open(bad_wizards, 'a') as f:
                    f.write(wizard_worker + '\n')
            return False, wizard_worker
        else:
            if not self.opt['is_sandbox']:
                with open(good_wizards, 'a') as f:
                    f.write(wizard_worker + '\n')
            return True, wizard_worker

    def review_work(self):
        global review_agent

        def review_agent(ag):
            role = ag.id
            for d in self.dialog:
                if role in d['speaker']:
                    if d['offensive']:
                        ag.reject_work(
                            reason='Your HIT has been rejected '
                            'because we detected offensive '
                            'language in your submission.'
                        )

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(review_agent)(agent) for agent in self.agents
        )

    def shutdown(self):
        """
        Shutdown all mturk agents in parallel, otherwise if one mturk agent is
        disconnected then it could prevent other mturk agents from completing.
        """
        global shutdown_agent

        def shutdown_agent(agent):
            agent.shutdown()

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(shutdown_agent)(agent) for agent in self.agents
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import boto3
import os
import json
import re
from datetime import datetime

from botocore.exceptions import ClientError
from botocore.exceptions import ProfileNotFound

region_name = 'us-east-1'
aws_profile_name = 'parlai_mturk'
client = None

parent_dir = os.path.dirname(os.path.abspath(__file__))


def setup_aws_credentials():
    try:
        # Use existing credentials
        boto3.Session(profile_name=aws_profile_name)
    except ProfileNotFound:
        # Setup new credentials
        print(
            'AWS credentials not found. Please create an IAM user with '
            'programmatic access and AdministratorAccess policy at '
            'https://console.aws.amazon.com/iam/ (On the "Set permissions" '
            'page, choose "Attach existing policies directly" and then select '
            '"AdministratorAccess" policy). After creating the IAM user, '
            'please enter the user\'s Access Key ID and Secret Access '
            'Key below:'
        )
        aws_access_key_id = input('Access Key ID: ')
        aws_secret_access_key = input('Secret Access Key: ')
        if not os.path.exists(os.path.expanduser('~/.aws/')):
            os.makedirs(os.path.expanduser('~/.aws/'))
        aws_credentials_file_path = '~/.aws/credentials'
        aws_credentials_file_string = None
        expanded_aws_file_path = os.path.expanduser(aws_credentials_file_path)
        if os.path.exists(expanded_aws_file_path):
            with open(expanded_aws_file_path, 'r') as aws_credentials_file:
                aws_credentials_file_string = aws_credentials_file.read()
        with open(expanded_aws_file_path, 'a+') as aws_credentials_file:
            # Clean up file
            if aws_credentials_file_string:
                if aws_credentials_file_string.endswith("\n\n"):
                    pass
                elif aws_credentials_file_string.endswith("\n"):
                    aws_credentials_file.write("\n")
                else:
                    aws_credentials_file.write("\n\n")
            # Write login details
            aws_credentials_file.write('[{}]\n'.format(aws_profile_name))
            aws_credentials_file.write(
                'aws_access_key_id={}\n'.format(aws_access_key_id)
            )
            aws_credentials_file.write(
                'aws_secret_access_key={}\n'.format(aws_secret_access_key)
            )
        print(
            'AWS credentials successfully saved in {} file.\n'.format(
                aws_credentials_file_path
            )
        )
    os.environ['AWS_PROFILE'] = aws_profile_name


def calculate_mturk_cost(payment_opt):
    """
    MTurk Pricing: https://requester.mturk.com/pricing 20% fee on the reward and bonus
    amount (if any) you pay Workers. HITs with 10 or more assignments will be charged an
    additional 20% fee on the reward you pay Workers.

    Example payment_opt format for paying reward:
    {
        'type': 'reward',
        'num_total_assignments': 1,
        'reward': 0.05  # in dollars
    }

    Example payment_opt format for paying bonus:
    {
        'type': 'bonus',
        'amount': 1000  # in dollars
    }
    """
    total_cost = 0
    if payment_opt['type'] == 'reward':
        mult = 1.2
        total_cost = payment_opt['num_total_assignments'] * payment_opt['reward'] * mult
    elif payment_opt['type'] == 'bonus':
        total_cost = payment_opt['amount'] * 1.2
    return total_cost


def check_mturk_balance(balance_needed, is_sandbox):
    """
    Checks to see if there is at least balance_needed amount in the requester account,
    returns True if the balance is greater than balance_needed.
    """
    client = get_mturk_client(is_sandbox)

    # Test that you can connect to the API by checking your account balance
    # In Sandbox this always returns $10,000
    try:
        user_balance = float(client.get_account_balance()['AvailableBalance'])
    except ClientError as e:
        if e.response['Error']['Code'] == 'RequestError':
            print(
                'ERROR: To use the MTurk API, you will need an Amazon Web '
                'Services (AWS) Account. Your AWS account must be linked to '
                'your Amazon Mechanical Turk Account. Visit '
                'https://requestersandbox.mturk.com/developer to get started. '
                '(Note: if you have recently linked your account, please wait '
                'for a couple minutes before trying again.)\n'
            )
            quit()
        else:
            raise

    if user_balance < balance_needed:
        print(
            'You might not have enough money in your MTurk account. Please go '
            'to https://requester.mturk.com/account and increase your balance '
            'to at least ${}, and then try again.'.format(balance_needed)
        )
        return False
    else:
        return True


def create_hit_config(opt, task_description, unique_worker, is_sandbox):
    """
    Writes a HIT config to file.
    """
    mturk_submit_url = 'https://workersandbox.mturk.com/mturk/externalSubmit'
    if not is_sandbox:
        mturk_submit_url = 'https://www.mturk.com/mturk/externalSubmit'
    hit_config = {
        'task_description': task_description,
        'is_sandbox': is_sandbox,
        'mturk_submit_url': mturk_submit_url,
        'unique_worker': unique_worker,
        'frame_height': opt.get('frame_height', 650),
        'allow_reviews': opt.get('allow_reviews', False),
        'block_mobile': opt.get('block_mobile', True),
        # Populate the chat pane title from chat_title, defaulting to the
        # hit_title if the task provides no chat_title
        'chat_title': opt.get('chat_title', opt.get('hit_title', 'Live Chat')),
        'template_type': opt.get('frontend_template_type', 'default'),
    }
    hit_config_file_path = os.path.join(opt['tmp_dir'], 'hit_config.json')
    if os.path.exists(hit_config_file_path):
        os.remove(hit_config_file_path)
    with open(hit_config_file_path, 'w') as hit_config_file:
        hit_config_file.write(json.dumps(hit_config))


def get_mturk_client(is_sandbox):
    """
    Returns the appropriate mturk client given sandbox option.
    """
    global client
    if client is None:
        client = boto3.client(
            service_name='mturk',
            region_name='us-east-1',
            endpoint_url='https://mturk-requester-sandbox.us-east-1.amazonaws.com',
        )
        # Region is always us-east-1
        if not is_sandbox:
            client = boto3.client(service_name='mturk', region_name='us-east-1')
    return client


def delete_qualification(qualification_id, is_sandbox):
    """
    Deletes a qualification by id.
    """
    client = get_mturk_client(is_sandbox)
    client.delete_qualification_type(QualificationTypeId=qualification_id)


def find_qualification(qualification_name, is_sandbox, must_be_owned=True):
    """
    Query amazon to find the existing qualification name, return the Id, otherwise
    return none.

    If must_be_owned is true, it only returns qualifications owned by the user. Will
    return False if it finds another's qualification
    """
    client = get_mturk_client(is_sandbox)

    # Search for the qualification owned by the current user
    response = client.list_qualification_types(
        Query=qualification_name, MustBeRequestable=True, MustBeOwnedByCaller=True
    )

    for qualification in response['QualificationTypes']:
        if qualification['Name'] == qualification_name:
            return qualification['QualificationTypeId']

    # Qualification was not found to exist, check to see if someone else has it
    response = client.list_qualification_types(
        Query=qualification_name, MustBeRequestable=True, MustBeOwnedByCaller=False
    )

    for qualification in response['QualificationTypes']:
        if qualification['Name'] == qualification_name:
            if must_be_owned:
                print(
                    'Sorry, the qualification name {} is already owned, '
                    'please use a different name for your qualification.'
                    ''.format(qualification_name)
                )
                return False
            return qualification['QualificationTypeId']
    return None


def find_or_create_qualification(
    qualification_name, description, is_sandbox, must_be_owned=True
):
    """
    Query amazon to find the existing qualification name, return the Id.

    If it exists and must_be_owned is true but we don't own it, this prints an error and
    returns none. If it doesn't exist, the qualification is created
    """
    qual_id = find_qualification(
        qualification_name, is_sandbox, must_be_owned=must_be_owned
    )

    if qual_id is False:
        return None
    if qual_id is not None:
        return qual_id

    # Create the qualification, as it doesn't exist yet
    client = get_mturk_client(is_sandbox)
    response = client.create_qualification_type(
        Name=qualification_name,
        Description=description,
        QualificationTypeStatus='Active',
    )
    return response['QualificationType']['QualificationTypeId']


def give_worker_qualification(worker_id, qualification_id, value=None, is_sandbox=True):
    """
    Give a qualification to the given worker.
    """
    client = get_mturk_client(is_sandbox)

    if value is not None:
        client.associate_qualification_with_worker(
            QualificationTypeId=qualification_id,
            WorkerId=worker_id,
            IntegerValue=value,
            SendNotification=False,
        )
    else:
        client.associate_qualification_with_worker(
            QualificationTypeId=qualification_id,
            WorkerId=worker_id,
            IntegerValue=1,
            SendNotification=False,
        )


def remove_worker_qualification(
    worker_id, qualification_id, is_sandbox=True, reason=''
):
    """
    Give a qualification to the given worker.
    """
    client = get_mturk_client(is_sandbox)
    client.disassociate_qualification_from_worker(
        QualificationTypeId=qualification_id, WorkerId=worker_id, Reason=reason
    )


def create_hit_type(
    hit_title,
    hit_description,
    hit_keywords,
    hit_reward,
    assignment_duration_in_seconds,
    is_sandbox,
    qualifications=None,
    auto_approve_delay=7 * 24 * 3600,  # default 1 week
):
    """
    Create a HIT type to be used to generate HITs of the requested params.
    """
    client = get_mturk_client(is_sandbox)

    # If the user hasn't specified a location qualification, we assume to
    # restrict the HIT to some english-speaking countries.
    locale_requirements = []
    has_locale_qual = False
    if qualifications is not None:
        for q in qualifications:
            if q['QualificationTypeId'] == '00000000000000000071':
                has_locale_qual = True
        locale_requirements += qualifications

    if not has_locale_qual:
        locale_requirements.append(
            {
                'QualificationTypeId': '00000000000000000071',
                'Comparator': 'In',
                'LocaleValues': [
                    {'Country': 'US'},
                    {'Country': 'CA'},
                    {'Country': 'GB'},
                    {'Country': 'AU'},
                    {'Country': 'NZ'},
                ],
                'RequiredToPreview': True,
            }
        )

    # Create the HIT type
    response = client.create_hit_type(
        AutoApprovalDelayInSeconds=auto_approve_delay,
        AssignmentDurationInSeconds=assignment_duration_in_seconds,
        Reward=str(hit_reward),
        Title=hit_title,
        Keywords=hit_keywords,
        Description=hit_description,
        QualificationRequirements=locale_requirements,
    )
    hit_type_id = response['HITTypeId']
    return hit_type_id


def create_hit_with_hit_type(opt, page_url, hit_type_id, num_assignments, is_sandbox):
    """
    Creates the actual HIT given the type and page to direct clients to.
    """
    page_url = page_url.replace('&', '&amp;')
    amazon_ext_url = (
        'http://mechanicalturk.amazonaws.com/'
        'AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd'
    )
    question_data_struture = (
        '<ExternalQuestion xmlns="{}">'
        '<ExternalURL>{}</ExternalURL>'  # noqa: E131
        '<FrameHeight>{}</FrameHeight>'
        '</ExternalQuestion>'
        ''.format(amazon_ext_url, page_url, opt.get('frame_height', 650))
    )

    client = get_mturk_client(is_sandbox)

    # Create the HIT
    response = client.create_hit_with_hit_type(
        HITTypeId=hit_type_id,
        MaxAssignments=num_assignments,
        LifetimeInSeconds=31536000,
        Question=question_data_struture,
    )

    # The response included several fields that will be helpful later
    hit_type_id = response['HIT']['HITTypeId']
    hit_id = response['HIT']['HITId']

    # Construct the hit URL
    url_target = 'workersandbox'
    if not is_sandbox:
        url_target = 'www'
    hit_link = 'https://{}.mturk.com/mturk/preview?groupId={}'.format(
        url_target, hit_type_id
    )
    return hit_link, hit_id, response


def expire_hit(is_sandbox, hit_id):
    client = get_mturk_client(is_sandbox)
    # Update expiration to a time in the past, the HIT expires instantly
    past_time = datetime(2015, 1, 1)
    client.update_expiration_for_hit(HITId=hit_id, ExpireAt=past_time)


def setup_sns_topic(task_name, server_url, task_group_id):
    # Create the topic and subscribe to it so that our server receives notifs
    client = boto3.client('sns', region_name='us-east-1')
    pattern = re.compile('[^a-zA-Z0-9_-]+')
    filtered_task_name = pattern.sub('', task_name)
    response = client.create_topic(Name=filtered_task_name)
    arn = response['TopicArn']
    topic_sub_url = '{}/sns_posts?task_group_id={}'.format(server_url, task_group_id)
    client.subscribe(TopicArn=arn, Protocol='https', Endpoint=topic_sub_url)
    response = client.get_topic_attributes(TopicArn=arn)
    policy_json = '''{{
    "Version": "2008-10-17",
    "Id": "{}/MTurkOnlyPolicy",
    "Statement": [
        {{
            "Sid": "MTurkOnlyPolicy",
            "Effect": "Allow",
            "Principal": {{
                "Service": "mturk-requester.amazonaws.com"
            }},
            "Action": "SNS:Publish",
            "Resource": "{}"
        }}
    ]}}'''.format(
        arn, arn
    )
    client.set_topic_attributes(
        TopicArn=arn, AttributeName='Policy', AttributeValue=policy_json
    )
    return arn


def subscribe_to_hits(hit_type_id, is_sandbox, sns_arn):
    # Get the mturk client and create notifications for our hits
    client = get_mturk_client(is_sandbox)
    client.update_notification_settings(
        HITTypeId=hit_type_id,
        Notification={
            'Destination': sns_arn,
            'Transport': 'SNS',
            'Version': '2006-05-05',
            'EventTypes': [
                'AssignmentAbandoned',
                'AssignmentReturned',
                'AssignmentSubmitted',
            ],
        },
        Active=True,
    )


def send_test_notif(topic_arn, event_type):
    client = get_mturk_client(True)
    client.send_test_event_notification(
        Notification={
            'Destination': topic_arn,
            'Transport': 'SNS',
            'Version': '2006-05-05',
            'EventTypes': [
                'AssignmentAbandoned',
                'AssignmentReturned',
                'AssignmentSubmitted',
            ],
        },
        TestEventType=event_type,
    )


def delete_sns_topic(topic_arn):
    client = boto3.client('sns', region_name='us-east-1')
    client.delete_topic(TopicArn=topic_arn)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.tasks.dealnodeal.worlds import MTurkDealNoDealDialogWorld
from parlai.agents.local_human.local_human import LocalHumanAgent
from parlai.core.agents import create_agent
from task_config import task_config


def main():
    """
    This task consists of one agent, model or MTurk worker, talking to an MTurk worker
    to negotiate a deal.
    """
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    argparser.add_argument(
        '--two_mturk_agents',
        dest='two_mturk_agents',
        action='store_true',
        help='data collection mode ' 'with converations between two MTurk agents',
    )

    opt = argparser.parse_args()
    opt['task'] = 'dealnodeal'
    opt['datatype'] = 'valid'
    opt.update(task_config)

    local_agent_1_id = 'local_1'
    mturk_agent_ids = ['mturk_agent_1']
    if opt['two_mturk_agents']:
        mturk_agent_ids.append('mturk_agent_2')

    mturk_manager = MTurkManager(opt=opt, mturk_agent_ids=mturk_agent_ids)

    mturk_manager.setup_server()

    try:
        mturk_manager.start_new_run()
        mturk_manager.create_hits()

        mturk_manager.set_onboard_function(onboard_function=None)
        mturk_manager.ready_to_accept_workers()

        def check_worker_eligibility(worker):
            return True

        def assign_worker_roles(workers):
            for index, worker in enumerate(workers):
                worker.id = mturk_agent_ids[index % len(mturk_agent_ids)]

        def run_conversation(mturk_manager, opt, workers):
            agents = workers[:]

            # Create a local agent
            if not opt['two_mturk_agents']:
                if 'model' in opt:
                    local_agent = create_agent(opt)
                else:
                    local_agent = LocalHumanAgent(opt=None)

                local_agent.id = local_agent_1_id
                agents.append(local_agent)

            opt["batchindex"] = mturk_manager.started_conversations

            world = MTurkDealNoDealDialogWorld(opt=opt, agents=agents)

            while not world.episode_done():
                world.parley()

            world.shutdown()

        mturk_manager.start_task(
            eligibility_function=check_worker_eligibility,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )

    except BaseException:
        raise
    finally:
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Model Configuration file for Fast ACUTE Eval.

CONFIG: Dict[str, Dict]
    - maps ids to their appropriate options
    - for models, please only include options that you would specify on the command line
"""
import os
from typing import Dict

ROOT_DIR = '/checkpoint/parlai/acute_evals/'
CONFIG: Dict[str, Dict] = {
    'example_model_1': {
        'model_file': 'zoo:tutorial_transformer_generator/model',
        'model': 'transformer/generator',
        # general args
        'batchsize': 1,
        'skip_generation': False,
        'interactive_mode': False,
        'beam_size': 3,
        'beam_min_length': 3,
        'inference': 'beam',
        'beam_block_ngram': 3,
        'beam_context_block_ngram': 3,
    },
    'example_model_2': {
        'model_file': 'zoo:tutorial_transformer_generator/model',
        'model': 'transformer/generator',
        # general args
        'batchsize': 1,
        'skip_generation': False,
        'interactive_mode': False,
        'inference': 'nucleus',
        'topp': 0.9,
    },
    'example_model_log': {
        'log_path': f"{os.path.dirname(os.path.realpath(__file__))}/example/chat_log.jsonl"
    },
    'example_dataset': {'task': 'convai2', 'prepended_context': True},
}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BART: Denoising Sequence-to-Sequence Pre-training for
Natural Language Generation, Translation, and Comprehension

See https://arxiv.org/abs/1910.13461.

The BART agent can be instantiated as simply `-m bart`,
however it is recommended to specify `--init-model zoo:bart/bart_large/model`
or `-mf zoo:bart/bart_large/model` to ensure correct dictionaries are saved.
"""
import os
import torch
from typing import Optional, Dict, Any

from parlai.agents.bart.convert_fairseq_to_parlai import ConversionScript
from parlai.agents.bart.modules import BartModel
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.agents import compare_init_model_opts
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import History
from parlai.core.torch_generator_agent import PPLMetric
from parlai.core.metrics import AverageMetric
from parlai.utils.typing import TShared
from parlai.zoo.bart.build import download, CONVERSION_ARGS, BART_ARGS


class BartAgent(TransformerGeneratorAgent):
    """
    BART Agent.

    Relies on the BART model implemented in fairseq.

    If you have a fine-tuned BART model from fairseq, you can specify the
    `--init-fairseq-model` arg, which will convert your fine-tuned model
    to a ParlAI model.
    """

    @staticmethod
    def add_cmdline_args(argparser: ParlaiParser):
        """
        Override to add init-fairseq-model arg.
        """
        TransformerGeneratorAgent.add_cmdline_args(argparser)
        group = argparser.add_argument_group('Bart Args')
        group.add_argument(
            '--init-fairseq-model',
            type=str,
            default=None,
            help='fairseq checkpoint for bart',
        )
        group.add_argument(
            '--output-conversion-path',
            type=str,
            default=None,
            help='where to save fairseq conversion',
        )
        argparser.set_defaults(dict_tokenizer='gpt2')

    def __init__(self, opt: Opt, shared: TShared = None):
        if not shared:
            opt = self._initialize_bart(opt)
        super().__init__(opt, shared)

    def _initialize_bart(self, opt: Opt) -> Opt:
        """
        Download and convert BART pre-trained models.

        Additionally, convert `init-fairseq-model` if necessary.

        :param opt:
            ParlAI-parsed options

        :return opt:
            return opt with BART-specific args.
        """
        if not opt.get('converting'):
            download(opt['datapath'])
            opt['init_model'] = os.path.join(
                opt['datapath'], 'models/bart/bart_large/model'
            )
        if opt.get('init_fairseq_model'):
            opt = self._convert_model(opt)
        opt.update(BART_ARGS)
        compare_init_model_opts(opt, opt)
        return opt

    def _get_conversion_args(self, opt: Opt) -> Dict[str, Any]:
        """
        Get args for fairseq model conversion.

        :param opt:
            ParlAI Opt

        :return args:
            returns dictionary of args to send to conversion script.
        """
        model_name = os.path.split(opt['init_fairseq_model'])[-1]
        args = CONVERSION_ARGS.copy()

        args['input'] = [opt['init_fairseq_model']]
        if opt.get('model_file') and not os.path.exists(opt['model_file']):
            args['output'] = opt['model_file']
        elif opt.get('output_conversion_path'):
            args['output'] = opt['output_conversion_path']
        else:
            args['output'] = os.path.join(
                opt['datapath'], 'models/converted_fairseq_models/', model_name
            )

        return args

    def _convert_model(self, opt: Opt) -> Opt:
        """
        Convert fairseq init model to ParlAI Model.

        :param opt:
            options

        :return opt:
            return opt with new init_model path
        """
        args = self._get_conversion_args(opt)
        ConversionScript.main(**args)
        opt['init_model'] = args['output']
        return opt

    def build_model(self) -> BartModel:
        """
        Build and return model.
        """
        model = BartModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    def _set_text_vec(
        self, obs: Message, history: History, truncate: Optional[int]
    ) -> Message:
        """
        Override to prepend start token and append end token.
        """
        obs = super()._set_text_vec(obs, history, truncate)
        if 'text' not in obs or 'text_vec' not in obs:
            return obs
        vec = obs['text_vec']
        if truncate is not None:
            vec = torch.LongTensor(  # type: ignore
                self._check_truncate(obs['text_vec'], truncate - 2, True)
            )
        obs.force_set(
            'text_vec', self._add_start_end_tokens(vec, add_start=True, add_end=True)
        )
        return obs

    def _get_initial_decoder_input(
        self, bsz: int, beam_size: int, dev: torch.device
    ) -> torch.LongTensor:
        """
        Override to seed decoder with EOS BOS token.
        """
        return (
            torch.LongTensor(  # type: ignore
                [self.END_IDX, self.START_IDX]
            )
            .expand(bsz * beam_size, 2)
            .to(dev)
        )

    def compute_loss(self, batch, return_output=False):
        """
        Override TGA.compute_loss to ignore start token.
        """
        if batch.label_vec is None:
            raise ValueError('Cannot compute loss without a label.')
        model_output = self.model(*self._model_input(batch), ys=batch.label_vec)
        scores, preds, *_ = model_output

        if scores.size(1) != batch.label_vec.size(1):
            # ignore start
            scores = scores[:, 1:, :]
            preds = preds[:, 1:]

        score_view = scores.reshape(-1, scores.size(-1))
        loss = self.criterion(score_view, batch.label_vec.view(-1))
        loss = loss.view(scores.shape[:-1]).sum(dim=1)
        # save loss to metrics
        notnull = batch.label_vec.ne(self.NULL_IDX)
        target_tokens = notnull.long().sum(dim=-1)
        correct = ((batch.label_vec == preds) * notnull).sum(dim=-1)

        self.record_local_metric('loss', AverageMetric.many(loss, target_tokens))
        self.record_local_metric('ppl', PPLMetric.many(loss, target_tokens))
        self.record_local_metric(
            'token_acc', AverageMetric.many(correct, target_tokens)
        )
        # actually do backwards loss
        loss = loss.sum()
        loss /= target_tokens.sum()  # average loss per token
        if return_output:
            return (loss, model_output)
        else:
            return loss

    def _construct_token_losses(self, labels, model_output):
        """
        Override TGA._construct_token_losses to ignore start token.
        """
        # Get non-aggregated losses
        scores, _, _ = model_output
        scores = scores[:, 1:, :]  # ignore start token
        score_view = scores.reshape(-1, scores.size(-1))
        losses = self.criterion(score_view, labels.view(-1)).view(len(labels), -1)

        # Zip decoded tokens with losses
        token_losses = []
        for i, label in enumerate(labels):
            token_losses.append(
                list(
                    zip(
                        [self.dict[token] for token in label.tolist()],
                        losses[i].tolist(),
                    )
                )
            )
        return token_losses



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://github.com/salesforce/WikiSQL/raw/master/data.tar.bz2',
        'data.tar.bz2',
        '755c728ab188e364575705c8641f3fafd86fb089cb8b08e8c03f01832aae0881',
    ),
    DownloadableFile(
        'https://github.com/salesforce/WikiSQL/raw/master/lib/query.py',
        'query.py',
        'f539150bea6cd07a5dca226abcced6f9d356d216f5c3d70107693613f1fbeb25',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'WikiSQL')
    version = 'None'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import os
from parlai.core.params import ParlaiParser
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.tasks.multi_agent_dialog.worlds import (
    MTurkMultiAgentDialogWorld,
    MTurkMultiAgentDialogOnboardWorld,
)
from parlai.agents.local_human.local_human import LocalHumanAgent
from parlai.mturk.tasks.multi_agent_dialog.task_config import task_config


def main():
    """
    This task consists of one local human agent and two MTurk agents, each MTurk agent
    will go through the onboarding step to provide information about themselves, before
    being put into a conversation.

    You can end the conversation by sending a message ending with `[DONE]` from human_1.
    """
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    opt = argparser.parse_args()
    opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))
    opt.update(task_config)

    mturk_agent_1_id = 'mturk_agent_1'
    mturk_agent_2_id = 'mturk_agent_2'
    human_agent_1_id = 'human_1'
    mturk_agent_ids = [mturk_agent_1_id, mturk_agent_2_id]
    mturk_manager = MTurkManager(opt=opt, mturk_agent_ids=mturk_agent_ids)
    mturk_manager.setup_server()

    try:
        mturk_manager.start_new_run()
        mturk_manager.create_hits()

        def run_onboard(worker):
            world = MTurkMultiAgentDialogOnboardWorld(opt=opt, mturk_agent=worker)
            while not world.episode_done():
                world.parley()
            world.shutdown()

        # You can set onboard_function to None to skip onboarding
        mturk_manager.set_onboard_function(onboard_function=run_onboard)
        mturk_manager.ready_to_accept_workers()

        def check_worker_eligibility(worker):
            return True

        eligibility_function = {'func': check_worker_eligibility, 'multiple': False}

        def assign_worker_roles(workers):
            for index, worker in enumerate(workers):
                worker.id = mturk_agent_ids[index % len(mturk_agent_ids)]

        def run_conversation(mturk_manager, opt, workers):
            # Create mturk agents
            mturk_agent_1 = workers[0]
            mturk_agent_2 = workers[1]

            # Create the local human agents
            human_agent_1 = LocalHumanAgent(opt={})
            human_agent_1.id = human_agent_1_id

            world = MTurkMultiAgentDialogWorld(
                opt=opt, agents=[human_agent_1, mturk_agent_1, mturk_agent_2]
            )

            while not world.episode_done():
                world.parley()

            world.shutdown()

        mturk_manager.start_task(
            eligibility_function=eligibility_function,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )

    except BaseException:
        raise
    finally:
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

task_config = {}

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Negotiate a deal with another user!'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config['hit_description'] = (
    'You will have a conversation with another user to agree how to divide some objects'
    'between you. Negotiate hard to get a deal worth as many points as possible!'
)


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,dialog'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''
<div id="preview">
You will have a conversation with another user to agree how to divide some objects between you. Negotiate hard to
get a deal worth as many points as possible<br><br>

If you are ready, please click "Accept HIT" to start this task.
</div>

<div id="input"></div>
<div float="right">
  <button class="btn btn-primary" style="width: 120px; font-size: 16px; float: left; margin-left: 10px; padding: 0px;" id="id_no_deal_button">No Agreement</button>
  <button class="btn btn-primary" style="width: 120px; font-size: 16px; float: left; margin-left: 10px; padding: 0px;" id="id_send_deal_button">Deal Agreed</button>
</div>

<script type="text/javascript">

var values = [0, 0, 0];
var counts = [0, 0, 0];

var num_messages = 0;
var numberOfItemTypes;

var your_selection = "";
var their_selection = "";

var image_path = "https://github.com/facebookresearch/end-to-end-negotiator/raw/master/src/images/";
var image_names = ["book", "hat", "ball"];

$("button#id_send_deal_button").hide();
$("button#id_no_deal_button").hide();

function enable_button(button, enable) {
    if (enable) {
      button.removeClass('disabled');
      button.prop("disabled", false);
    } else {
      button.addClass("disabled");
      button.prop("disabled", true);
    }
}

function makeInput(items) {
  $('#preview').html("");

  var table = $('#input');
    table.append('<h1>Divide these items between you and your partner. </h1>');
    table.append('<p><i>Your partner sees the <b>SAME ITEMS</b> but with <b>DIFFERENT VALUES</b></i></p>');
    table.append('<p><i>You get some items, and your partner will get the rest</i></p>');
    table.append('<p><i>Please try hard to negotiate a good deal for you!</i></p><hr>');
    table.append('<table>');
    table.append('<tr><td colspan="4"><b>Items you BOTH see</b></td><td style="padding:0 15px 0 15px;"><b>Value Each to <i>You</i></b></td><td><b>Number You Get</b></td></tr><br><hr>');
    var item = 0;

    items.forEach(count_value => {
        var number = count_value[0];
        var value = count_value[1];
        values[item] = value;
        counts[item] = number;

        var string = '';

        string += '<tr>';
        for (var i=0; i<4; i++) {
          string += '<td>';
            if (i < number) {
              string += '<img width=75px src="' + image_path + '/' + image_names[item] + '.png"></img>';
            }
          string += '</td>';
        }
        string += '<td align="center" valign="middle"><p style="font-size:20px">' + value + "</p></td>";
        string += '<td valign="middle">';
        string += '<select id="item' + item + '">';
          for (var i=0; i<=number; i++) {
            string += '<option value="' + i + '">You get ' + i + ', they get ' + (number - i) + '</option>';
          }
        string += '</select>';
        string += '</td></tr>';
        table.append(string);

        item++;

    });
    numberOfItemTypes = item;
    table.append('</table>');
    $("button#id_send_deal_button").show();
    enable_button($("button#id_send_deal_button"), false);
}

(function() {
    // Override handle_new_message function
    handle_new_message = function() {
        var new_message_id = arguments[0];
        var message = arguments[1];
        var agent_id = message.id;
        var text = message.text;
        var items = message.items;
        var was_this_agent = (agent_id == cur_agent_id);

        if (displayed_messages.indexOf(new_message_id) !== -1) {
            // This message has already been seen and put up into the chat
            log(new_message_id + ' was a repeat message', 1);
            return;
        }

        log('New message, ' + new_message_id + ' from agent ' + agent_id, 1);
        displayed_messages.push(new_message_id);

        if (message.items) {
            makeInput([[items.book_cnt, items.book_val], [items.hat_cnt, items.hat_val], [items.ball_cnt, items.ball_val]]);
        } else if (text.startsWith('<selection>')) {
            enable_button($("button#id_send_deal_button"), !was_this_agent && your_selection == "");
            enable_button($("button#id_no_deal_button"), !was_this_agent && your_selection == "");
            if (!was_this_agent) {
                if (your_selection == "") {
                    $("button#id_no_deal_button").show();
                }
                $('#response-type-text-input').html("Your partner said a deal was agreed. " +
                    "Please enter the the agreed deal, or 'no agreement' if you don't think you agreed a deal.");
                their_selection = text;
            } else {
                your_selection = text;
            }

            completion_message(your_selection, their_selection);
        } else {
            num_messages++;
            if (num_messages >= 2) {
                enable_button($("button#id_send_deal_button"), !was_this_agent);
            }

            if (num_messages >= 10) {
                $("button#id_no_deal_button").show();
                enable_button($("button#id_no_deal_button"), !was_this_agent);
            }

            message.id = (was_this_agent ? "YOU:" : "THEM:");

            var agent_id = message.id;

            add_message_to_conversation(was_this_agent ? "YOU" : "THEM", text, was_this_agent);
        }
    };
})();

function completion_message(your_selection, their_selection) {
    if (your_selection == "" || their_selection == "") {
        // Haven't both entered deal
        return ;
    }

    var your_counts = your_selection.match(/\d+/g);
    var their_counts = their_selection.match(/\d+/g);
    var agree = true;
    var all_zero = true;
    var score = 0;

    for (var i=0; i<counts.length; i++) {
        var your_count = parseInt(your_counts[2 * i + 1]);
        var their_count = parseInt(their_counts[2 * i + 1]);

        all_zero = all_zero && your_count == 0;
        all_zero = all_zero && their_count == 0;
        agree = agree && ((your_count + their_count) == counts[i]);
        score += your_count * values[i];
    }

    if (!agree) {
        score = 0;
    }

    var msg = "";
    msg += "<br><h1> Your score: " + score + "/10</h1>";

    if (all_zero) {
        msg += "Please try to reach agreements with your partner in future.";
    } else if (!agree) {
        msg += "You and your partner entered <b>different deals</b>. " +
                "Please try to carefully agree deals in future, or your work may be rejected.";
    } else if (score < 5) {
        msg += "Please <b>negotiate harder</b> in future to get good deals or your work may be rejected.";
    } else if (score < 7 && num_messages < 4) {
        msg += "Please <b>negotiate harder</b> in future to get good deals or your work may be rejected.";
    } else if (score == 9 || score == 10) {
        msg += "You got a <b>great deal</b>, keep up the good work!";
    } else {
        msg += "Thanks for successfully agreeing a deal.";
    }

    $('#input').html(msg);

}


$("button#id_no_deal_button").on('click', function () {
    send_deal("<selection> item0=0 item1=0 item2=0");
});


function send_deal(selection) {
      // Disable the send button
      enable_button($("button#id_no_deal_button"), false);
      enable_button($("button#id_send_msg_button"), false);
      enable_button($("button#id_send_deal_button"), false);

      new_message_id = uuidv4();
      your_selection = selection;
      completion_message(your_selection, their_selection);
      if (!their_selection) {
        $('#response-type-text-input').html("Waiting for your partner to enter the deal...");
      }

      // Send a packet
      send_packet(
        TYPE_MESSAGE,
        {
          text: selection,
          id: cur_agent_id,
          message_id: new_message_id,
          episode_done: false
        },
        true,
        true,
        function(msg) {
        }
      );
}

$("button#id_send_deal_button").on('click', function () {
  var score = 0;
  var selection = '<selection>';
  var max = 0;
  var num_items = 0;

  for (var i=0; i<counts.length; i++) {
    var e = document.getElementById("item" + i);
    var num = e.options[e.selectedIndex].value;
    score += values[i] * num;
    max += values[i] * counts[i];
    selection += ' item' + i + '=' + num;
    num_items += counts[i];
  }

  if (num_items == 0) {
    return ;
  }

  send_deal(selection);
});

</script>
'''



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

task_config = {}

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Chat with and rate a fellow conversationalist!'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config['hit_description'] = (
    'Choose a topic and chat with another user about it. '
    + 'After the conversation, you will be asked to rate your partner.'
)


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,dialog'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = """
<br>
<b><h2>Task Description</h2></b>
<br>
In this task, you will have a conversation with another user.
The goal of this task is to go in depth about something that interests you,
while keeping the conversation <b><span style="color:red">engaging</span></b>
and <b><span style="color:red">fun</span></b>.
<br>
<br>
Either you or your partner will choose between 2-3 topics to discuss.
<b>NOTE: If you are the one choosing a topic, you do not need to have prior
knowledge of the topic you choose</b>. Rather, it could be something you'd like
to learn more about.
<br>
<br>
<b>If you complete the task, you will receive $1.00</b>. It may take up to
24 hours to review the HITs, so please allow that much time to pass before
payment. Please note that you may only complete <b>one</b> of these HITs.
After completion, you will be assigned a qualification that prevents you
from working on more.
<br>
<br>
After a given number of turns, you may be asked a few questions in order to
evaluate your partner.
<br>
<br>
<b>If your partner answers poorly, change topic.</b> Do not linger on their
poor response. Instead, mention this during the evaluation.
<br>
<br>
<h4><span style="color:blue"><b>Close Window/Timeout/Return HIT</b></span></h4>
Once the conversation has started, close window/timeout or return HIT during
the chat will result in <b>HIT EXPIRED</b> to you and NO reward paid.
<br>
<br>
<h4><span style="color:blue"><b>Important Notice</b></span></h4>
1. <b>Be aware the conversations you have will be made public, so act as you
would e.g. on a public social network like Twitter.</b>
<br>
2. Please do not send long messages: messages cannot exceed 30 words.
<br>
3. Please do not reference the task or MTurk itself during the conversation,
but speak naturally to the other person.
<br>
4. Please do not send any message that could make others uncomfortable,
including any level of discrimination, racism, sexism and offensive
religious/politics comments, otherwise the submission will be rejected.
<br>
<br>
Note: the user you are chatting with may be a human or a bot.
"""



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Log metrics to tensorboard.

This file provides interface to log any metrics in tensorboard, could be
extended to any other tool like visdom.

.. code-block: none

   tensorboard --logdir <PARLAI_DATA/tensorboard> --port 8888.
"""

import json
import numbers
from parlai.core.opt import Opt
from parlai.core.metrics import Metric
from parlai.utils.io import PathManager
import parlai.utils.logging as logging


class TensorboardLogger(object):
    """
    Log objects to tensorboard.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        """
        Add tensorboard CLI args.
        """
        logger = argparser.add_argument_group('Tensorboard Arguments')
        logger.add_argument(
            '-tblog',
            '--tensorboard-log',
            type='bool',
            default=False,
            help="Tensorboard logging of metrics, default is %(default)s",
            hidden=False,
        )

    def __init__(self, opt: Opt):
        try:
            # tensorboard is a very expensive thing to import. Wait until the
            # last second to import it.
            from tensorboardX import SummaryWriter
        except ImportError:
            raise ImportError('Please run `pip install tensorboard tensorboardX`.')

        tbpath = opt['model_file'] + '.tensorboard'
        logging.debug(f'Saving tensorboard logs to: {tbpath}')
        if not PathManager.exists(tbpath):
            PathManager.makedirs(tbpath)
        self.writer = SummaryWriter(tbpath, comment=json.dumps(opt))

    def log_metrics(self, setting, step, report):
        """
        Add all metrics from tensorboard_metrics opt key.

        :param setting:
            One of train/valid/test. Will be used as the title for the graph.
        :param step:
            Number of parleys
        :param report:
            The report to log
        """
        for k, v in report.items():
            if isinstance(v, numbers.Number):
                self.writer.add_scalar(f'{k}/{setting}', v, global_step=step)
            elif isinstance(v, Metric):
                self.writer.add_scalar(f'{k}/{setting}', v.value(), global_step=step)
            else:
                logging.error(f'k {k} v {v} is not a number')

    def flush(self):
        self.writer.flush()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from argparse import ArgumentParser

from parlai.projects.self_feeding.utils import extract_parlai_episodes
from parlai.mturk.tasks.self_feeding.rating.worlds import (
    NEW_TOPIC_REQUEST,
    SUGGESTION_REQUEST,
)
from parlai.utils.io import PathManager

# Initial prompts vary due to the random nouns, but all will start this way
INITIAL_PROMPT = "start a conversation"


def setup_args():
    argparser = ArgumentParser()
    argparser.add_argument('-if', '--infile', type=str)
    argparser.add_argument('-of', '--outfile', type=str)
    config = vars(argparser.parse_args())

    return config


def main(config):
    """
    Creates input files for y_exp mturk task from conversation/rating mturk task.

    input: file of logs (in ParlaiDialog format) from Mturk task 1 with turn-by-turn
        quality ratings 1-5
    output: file of logs (in ParlaiDialog format) sliced up to begin at the start of
        an episode or following a new topic request, and ending with a y_exp
    """
    new_episodes = []
    old_episodes = [e for e in extract_parlai_episodes(config['infile'])]
    for episode in old_episodes:
        for parley in episode:
            if any(
                parley.context.startswith(x)
                for x in (NEW_TOPIC_REQUEST.lower(), INITIAL_PROMPT.lower())
            ):
                new_episode = []
            new_episode.append(parley)
            if parley.context.startswith(SUGGESTION_REQUEST.lower()):
                new_episodes.append(new_episode)

    # Create parlai dialog file for easy viewing
    with PathManager.open(config['outfile'], 'w') as f:
        for episode in new_episodes:
            num_parleys = len(episode)
            for i, parley in enumerate(episode):
                if i == num_parleys - 1:
                    parley.episode_done = True
                f.write(f"{i}\t{parley.to_parlai()}\n")
    print(
        f"Extracted {len(new_episodes)} episodes out of {len(old_episodes)} "
        f"original episodes and wrote them to {config['outfile']}."
    )


if __name__ == '__main__':
    config = setup_args()
    main(config)
